from math import floor, ceil
import numpy as np

from ._base import Resource
from ._utils import (
    get_element_format_from_numpy_array,
    calculate_buffer_chunk_size,
    get_merged_blocks_from_mask_1d,
    check_data_is_clean_for_performance,
    is_little_endian,
    make_little_endian,
    logger,
)


class Buffer(Resource):
    """The Buffer represents a contiguous piece of GPU memory.

    A buffer can be used as index buffer or storage buffer. They are also used
    for uniform buffers (internally in the pygfx materials). You can provide
    (and update data for it), or use it as a placeholder for a buffer with no
    representation on the CPU.

    Parameters
    ----------
    data : array | None
        The initial data of the buffer. It must support the buffer-protocol,
        (e.g. a bytes or numpy array). If None, ``nbytes`` and ``nitems`` must
        be provided. The data will be accessible at ``buffer.data``, no copies
        are made.
    nbytes : int | None
        The size of the buffer in bytes. If both ``data`` and ``nbytes`` are
        given, the ``data.nbytes`` is checked against ``nbytes``.
    nitems : int | None
        The number of elements in the buffer. If both ``data`` and ``nitems``
        are given, the data is interpreted as having that many items (reshaped
        internally).
    format : None | str | ElementFormat | wgpu.VertexFormat | wgpu.IndexFormat
        A format string describing the buffer layout. This can follow pygfx'
        ``ElementFormat`` e.g. "3xf4", or wgpu's ``VertexFormat``. Optional: if
        None, it is automatically determined from the data.
    chunk_size : None | int
        The chunk size to use for uploading data to the GPU, expressed in items
        counts. When None (default) an optimal chunk size is determined
        automatically.
    force_contiguous : bool
        When set to true, the buffer goes into a stricter mode, forcing set data
        to be c_contiguous. This ensures optimal upload performance for cases
        when the data changes often.
    usage : int | wgpu.BufferUsage
        The wgpu ``usage`` flag for this buffer. Optional: typically pygfx can
        derive how the buffer is used and apply the appropriate flag. In cases
        where it doesn't this param provides an override. This is a bitmask flag
        (values are OR'd).

    Performance tips:

    * If the given data is not c_contiguous, the chunks will need to be copied
      at upload-time, which reduces performance when the data is changed often.
    * Setting ``force_contiguous`` ensures that the set data is contiguous, it
      is recommended to use this when the bufer data is dynamic.
    """

    def __init__(
        self,
        data=None,
        *,
        nbytes=None,
        nitems=None,
        format=None,
        chunk_size=None,
        force_contiguous=False,
        usage=0,
    ):
        super().__init__()
        Resource._rev += 1
        self._rev = Resource._rev

        # Attributes for internal use, updated by other parts of pygfx.
        self._wgpu_object = None
        self._wgpu_usage = int(usage)

        # Init
        self._data = None
        self._view = None
        self._force_contiguous = bool(force_contiguous)

        # Process data
        if data is not None:
            # Store data and view, and do some basic checks.
            # The view is a numpy array, but we go via memoryview to ensure data follows the buffer protocol.
            self._data = data
            self._view = view = np.asarray(memoryview(data))
            if self._force_contiguous:
                check_data_is_clean_for_performance("buffer", view)
            the_nbytes = view.nbytes
            if nbytes is not None and int(nbytes) != the_nbytes:
                raise ValueError("Given nbytes does not match size of given data.")
            # Establish number of items
            if nitems is not None:
                the_nitems = int(nitems)
            elif view.shape:
                the_nitems = view.shape[0]
            else:
                the_nitems = 1  # A scalar, e.g. a uniform struct
            reshape_array(view, the_nitems)
            # Establish format
            detected_format = None
            element_format = get_element_format_from_numpy_array(view)
            if element_format:
                elements_per_item = int(np.prod(view.shape[1:], initial=1))
                detected_format = (f"{elements_per_item}x" + element_format).lstrip(
                    "1x"
                )
        elif nbytes is not None and nitems is not None:
            # No data on the CPU side
            the_nbytes = int(nbytes)
            the_nitems = int(nitems)
            # Check
            if the_nbytes <= 0 or the_nitems <= 0:
                raise ValueError("Buffer size cannot be zero")
            bytes_per_item = the_nbytes // the_nitems
            if bytes_per_item * the_nitems != the_nbytes:
                raise ValueError("The given nbytes is not a multiple of nitems.")
            detected_format = None
        else:
            raise ValueError(
                "Buffer must be instantiated with either data or nbytes and nitems."
            )

        # Check size
        if the_nitems == 0:
            raise ValueError("Buffer size cannot be zero.")

        # Store derived props
        self._store.nbytes = the_nbytes
        self._store.nitems = the_nitems
        if format is not None:
            self._store.format = str(format)
        elif detected_format:
            self._store.format = detected_format
        else:
            self._store.format = None

        # Can now init other properties
        self.draw_range = 0, the_nitems

        # Get optimal chunk size
        if chunk_size is None:
            chunk_size = calculate_buffer_chunk_size(
                the_nitems,
                bytes_per_element=the_nbytes // the_nitems,
                byte_align=16,
                target_chunk_count=32,
            )
        else:
            chunk_size = min(max(int(chunk_size), 1), the_nitems)

        # Init chunks map
        if data is None:
            self._chunks_dirt_flag = 0
            self._chunk_size = 0
            self._chunk_mask = None
            self._chunk_list = []
        else:
            self._chunks_dirt_flag = 2
            self._chunk_size = chunk_size
            n_chunks = ceil(the_nitems / self._chunk_size)
            self._chunk_mask = np.ones((n_chunks,), bool)
            self._chunk_list = None

    @property
    def data(self):
        """The data for this buffer.

        Can be None if the data only exists on the GPU. This object is the same
        that was given to instantiate this object or with ``set_data()``.
        """
        return self._data

    @property
    def view(self):
        """A numpy array view on the data of this buffer.

        Can be None if the data only exists on the GPU. This is a view on the
        same memory as ``.data``. The first dimension matches ``nitems``.
        """
        return self._view

    @property
    def nbytes(self):
        """The number of bytes in the buffer."""
        # Note: many properties are stored on ._store, even if they cannot
        # change. This is done so that whan a buffer is swapped from another, we
        # can track what properties effectively changed. E.g. to determine
        # whether a shader recompilation is necessary.
        return self._store.nbytes

    @property
    def nitems(self):
        """The number of items in the buffer."""
        return self._store.nitems

    @property
    def itemsize(self):
        """The number of bytes for a single item."""
        # Note: For regular NxM buffers this can also be calculated from the
        # format, but not when the format is more complex / None, as with
        # uniform buffers (structured arrays).
        nbytes = self._store.nbytes  # deliberately touch
        nitems = self._store.nitems  # deliberately touch
        if nitems > 0:
            return nbytes // nitems
        elif self._view is not None:
            nelements_per_item = int(np.prod(self._view.shape[1:], initial=1))
            return nelements_per_item * self._view.itemsize
        else:
            raise RuntimeError("Cannot determine Buffer.itemsize")

    @property
    def format(self):
        """The buffer format.

        Usually a pygfx format specifier (e.g. 'u2' for scalar uint16, or '3xf4'
        for 3xfloat32), but can also be a value from ``wgpu.VertexFormat``, or
        None e.g. for uniform buffers.
        """
        return self._store.format

    @property
    def usage(self):
        """Bitmask indicating how the buffer can be used in a wgpu pipeline."""
        return self._wgpu_usage

    @property
    def vertex_byte_range(self):
        raise DeprecationWarning(
            "vertex_byte_range is deprecated, use draw_range instead."
        )

    @vertex_byte_range.setter
    def vertex_byte_range(self, offset_nbytes):
        raise DeprecationWarning(
            "vertex_byte_range is deprecated, use draw_range instead."
        )

    @property
    def draw_range(self):
        """The range to data (origin, size) expressed in items."""
        return self._store.draw_range

    @draw_range.setter
    def draw_range(self, draw_range):
        if draw_range is None:
            draw_range = 0, self.nitems
        origin, size = draw_range
        origin, size = int(origin), int(size)
        if not (origin == 0 or 0 < origin < self.nitems):  # note nitems can be 0
            raise ValueError("draw_range origin out of bounds.")
        if not (size >= 0 and origin + size <= self.nitems):
            raise ValueError("draw_range size out of bounds.")
        self._store.draw_range = origin, size
        Resource._rev += 1
        self._rev = Resource._rev

    def send_data(self, offset, data):
        """Send a chunk of data to the GPU.

        This provides a way to upload data to buffers that don't have local
        data (i.e. ``buffer.data is None``). It is intended for use-cases where
        data-copies must be avoid for performance. Can only be used when the
        buffer has no local data, and requires ``usage=wgpu.BufferUsage.COPY_DST``.

        The offset is expressed as an integer number of items.
        Note that in contrast to the ``update_x`` methods, multiple calls are not
        combined; each call to ``send_data()`` results in one upload operation.

        Example:

        .. code-block:: py

            buf = gfx.Buffer(
                nitems=64,
                nbytes=64*4*3,
                format="3xf4",
                usage=wgpu.BufferUsage.COPY_DST,
                force_contiguous=True,
            )
        """
        if self._data is not None:
            raise RuntimeError(
                "Can only use buffer.send_data() if the buffer has no local data."
            )
        # Check input
        offset = int(offset)
        if offset < 0:
            raise ValueError("offset must be a positive int")
        # Get data size
        size = data.shape[0]
        if (offset + size) > self.nitems:
            raise ValueError("The data with this offset does not fit.")
        # Create chunk
        if not data.flags.c_contiguous:
            if self._force_contiguous:
                raise ValueError(
                    "When force_contiguous is set, data passed to send_data() must be contiguous."
                )
            data = np.ascontiguousarray(data)
        self._chunk_list.append((offset, size, data))
        # Request sync
        Resource._rev += 1
        self._rev = Resource._rev
        self._gfx_mark_for_sync()

    def set_data(self, data):
        """Reset the data to a new array.

        This avoids a data-copy compared to doing ``buffer.data[:] = new_data``.
        The new data must fit the buffer's shape and format.
        """
        # Get view
        view = np.asarray(memoryview(data))
        # Do couple of checks
        if self._force_contiguous:
            check_data_is_clean_for_performance("buffer", view)
        if view.nbytes != self._view.nbytes:
            raise ValueError("buffer.set_data() nbytes does not match.")
        if view.dtype != self._view.dtype:
            raise ValueError("buffer.set_data() format does not match.")
        # Make sure the shape is ok. We only care about the first dimension.
        reshape_array(view, self.nitems)
        # Ok
        self._data = data
        self._view = view
        self.update_full()

    def update_full(self):
        """Mark the whole data for upload."""
        if self._chunks_dirt_flag == 2:
            return
        # self._chunk_mask.fill(True)  # implicit
        self._chunks_dirt_flag = 2
        Resource._rev += 1
        self._rev = Resource._rev
        self._gfx_mark_for_sync()

    def update_indices(self, indices):
        """Mark specific item indices for upload."""
        if self._chunks_dirt_flag == 2:
            return
        indices = np.asarray(indices)
        div = self._chunk_size
        self._chunk_mask[indices // div] = True
        self._chunks_dirt_flag = 1
        Resource._rev += 1
        self._rev = Resource._rev
        self._gfx_mark_for_sync()

    def update_range(self, offset=0, size=None):
        """Mark a certain range of the data for upload to the GPU.

        The offset and size are expressed in integer number of items.
        """
        # See ThreeJS BufferAttribute.updateRange
        if self._chunks_dirt_flag == 2:
            return

        nitems = self.nitems
        # Normalize inputs
        offset = int(offset or 0)
        size = int(nitems if size is None else size)
        # Checks
        if size == 0:
            return
        elif size < 0:
            raise ValueError("Update size must not be negative")
        elif offset < 0:
            raise ValueError("Update offset must not be negative")
        # Get indices
        index1 = offset
        index2 = min(nitems, offset + size)
        # Shortcut?
        if index1 == 0 and index2 == nitems:
            return self.update_full()
        # Update map
        div = self._chunk_size
        self._chunk_mask[floor(index1 / div) : ceil(index2 / div)] = True
        self._chunks_dirt_flag = 1
        Resource._rev += 1
        self._rev = Resource._rev
        self._gfx_mark_for_sync()

    def _gfx_get_chunk_descriptions(self):
        """Get a list of (offset, size) tuples, that can be
        used in _gfx_get_chunk_data(). This method also clears
        the chunk dirty statuses.
        """

        # In no-local-data mode, we (only) have a chunk list
        if self._chunk_list:
            chunks = self._chunk_list
            self._chunk_list = []
            return chunks

        if not self._chunks_dirt_flag:
            return []
        elif self._chunks_dirt_flag == 2:
            chunk_descriptions = [(0, self.nitems)]
        elif np.all(self._chunk_mask):
            chunk_descriptions = [(0, self.nitems)]
        else:
            # Get merged chunk blocks, using a smart algorithm.
            chunk_blocks = get_merged_blocks_from_mask_1d(self._chunk_mask)

            # Turn into proper descriptions, with chunk indices/counts scaled with the chunk size.
            chunk_descriptions = []
            chunk_size = self._chunk_size
            nitems = self._store["nitems"]
            for x, nx in chunk_blocks:
                offset = x * chunk_size
                size = min(nx * chunk_size, nitems - offset)
                chunk_descriptions.append((offset, size))

        # Reset
        self._chunks_dirt_flag = 0
        self._chunk_mask.fill(False)
        return chunk_descriptions

    def _gfx_get_chunk_data(self, offset, size):
        """Return subdata as a contiguous array."""

        # Get chunk
        if offset == 0 and size == self.nitems:
            chunk = self._view
        else:
            chunk = self._view[offset : offset + size]

        # Normalize the chunk
        if not is_little_endian(chunk):
            chunk = make_little_endian(chunk)
            if self._force_contiguous:
                logger.warning(
                    "force_contiguous was set, but chunk data is still big endian"
                )
        elif not chunk.flags.c_contiguous:
            if self._force_contiguous:
                logger.warning(
                    "force_contiguous was set, but chunk data is still discontiguous"
                )
            chunk = np.ascontiguousarray(chunk)

        return chunk


def reshape_array(view, n):
    """Reshape array so it's shape[0] is n."""
    if not (view.shape and view.shape[0] == n):
        elements_per_item = -1
        if n == 0:
            elements_per_item = int(np.prod([max(i, 1) for i in view.shape], initial=1))
        # This can fail if the data is not contiguous and strides don't work out.
        view.shape = (n, elements_per_item)



"""Computational geometry.

This may at some pointe be moved to https://github.com/pygfx/pycompgeo.
"""


def get_visible_part_of_line_ndc(ndc1, ndc2):
    """Get the visible part of the line, given by two homogeneous ndc coords.

    Returns (t1, t2) representing the visible line. If the two values
    are equal, the line is not in the viewport.
    """

    # Uncomment for performance measurements
    # import time
    # time_0 = time.perf_counter()

    # Get closest t for all 4 edges
    tx1 = binary_search_for_ndc_edge(ndc1, ndc2, -1, 0)
    tx2 = binary_search_for_ndc_edge(ndc1, ndc2, +1, 0)
    ty1 = binary_search_for_ndc_edge(ndc1, ndc2, -1, 1)
    ty2 = binary_search_for_ndc_edge(ndc1, ndc2, +1, 1)

    # Sort (because line can be oriented in any way)
    tx1, tx2 = min(tx1, tx2), max(tx1, tx2)
    ty1, ty2 = min(ty1, ty2), max(ty1, ty2)

    # Combine dimensions
    t1, t2 = max(tx1, ty1), min(tx2, ty2)

    # Sort again
    t1, t2 = min(t1, t2), max(t1, t2)

    # print(f"{time.perf_counter() - time_0:f}")

    return t1, t2


def binary_search_for_ndc_edge(ndc1, ndc2, ref, dim, *, n_iters=10):
    """Get the t that results in the smallest distance to ref in dimension dim.

    The ndc1 and ndc2 arguments must be in homogeneous ndc coords, so
    that they can be interpolated. The resulting t represents the point
    on the line between ndc1 and ncd2.

    This algorithm operates either over x or y, depending on dim (0 or1).
    So in order to determine the visible piece of the line, this
    function must be called 4 times.
    """

    # This algorithm performs a binary search, so that we can cover a
    # fine grid without doing an exhaustive search.
    #
    # Note that (with a perspective camera) the distance to the
    # reference does not scale linearly with t. The consequence is that
    # there is not a closed form solution to this problem. However, for
    # an orthographic camera there is!
    #
    # The binary search is done by sampling three points, then determining
    # whether the ref value is in 0..1 or 1..2, and then sampling one new
    # value in between. Repeat.
    #
    #  |------|------|
    #  0      1      2

    # Reduce the ndc positions to two scalars per position
    x1, x2 = float(ndc1[dim]), float(ndc2[dim])
    w1, w2 = float(ndc1[3]), float(ndc2[3])

    # Get the starting t's.
    # Things get iffy when the w is zero or negative, so we first find
    # the range where w is positive.
    initial_t1, initial_t2 = 0.0, 1.0
    eps = 1e-8
    if w1 < eps or w2 < eps:
        if w1 < eps and w2 < eps:
            # Camera is perspective and the end-points are both behind the camera.
            return 0.0 if w1 >= w2 else 1.0
        elif w1 < eps:
            # The first point is behind, move it.
            initial_t1 = (eps - w1) / (w2 - w1)
        else:  # w2 < eps:
            # The second point is behind, move it.
            initial_t2 = 1.0 - (eps - w2) / (w1 - w2)

    # Function to get a value corresponding to a given t
    def value_from_t(t):
        x = x1 * (1 - t) + x2 * t
        w = w1 * (1 - t) + w2 * t
        return x / w

    # Produce 3 tuples (t, value) representing the relative t-values as shown in above ascii diagram
    samples_t = [initial_t1, 0.5 * (initial_t1 + initial_t2), initial_t2]
    samples_val = [value_from_t(t) for t in samples_t]

    # If the values are very close, the line could be a point, or orthogonal to this dim
    if abs(samples_val[0] - samples_val[2]) < 1e-9:
        return 1.0 if samples_val[0] < ref else 0.0

    # Determine function for bisect.
    values_are_ascending = samples_val[0] <= samples_val[2]
    bisect = bisect_asc if values_are_ascending else bisect_desc

    # Do the first bisection!
    i = bisect(samples_val, ref)

    # Go deeper, if we must
    if w1 == 1.0 and w2 == 1.0:
        pass  # Ortho camera
    elif i == 0 or i == 3:
        pass  # The whole line is on one side of the ref
    else:
        # Binary search. With 10 iters we cover over 1K samples.
        for _ in range(n_iters):
            if i <= 1:
                samples_t[2] = samples_t[1]
                samples_val[2] = samples_val[1]
            else:
                samples_t[0] = samples_t[1]
                samples_val[0] = samples_val[1]
            new_t = 0.5 * (samples_t[0] + samples_t[2])
            samples_t[1] = new_t
            samples_val[1] = value_from_t(new_t)
            i = bisect(samples_val, ref)

    # Fine tune using a linear fit
    t_step = samples_t[1] - samples_t[0]
    if i == 0:
        t = samples_t[0]
    elif i == 3:
        t = samples_t[2]
    elif i == 1:
        y1 = samples_val[0]
        y2 = samples_val[1]
        dt = (ref - y1) / (y2 - y1)
        t = samples_t[0] + dt * t_step
    else:  # i == 2
        y1 = samples_val[1]
        y2 = samples_val[2]
        dt = (ref - y1) / (y2 - y1)
        t = samples_t[1] + dt * t_step

    return t


# We use our own bisect functions. Otherwise, to use bisect.bisect with
# descending values, we need the key param, which adds complexity, but
# also does not work pre py 3.10. This solution even seems a wee bit
# faster than using builtin bisect.


def bisect_asc(values, ref):
    """Simple version of bisect.bisect(). Values are assumed to be in ascending order."""
    for i, v in enumerate(values):
        if v > ref:
            return i
    return i + 1


def bisect_desc(values, ref):
    """Like bisect(), but the values are assumed in descnding order."""
    for i, v in enumerate(values):
        if v <= ref:
            return i
    return i + 1



import re
import logging


logger = logging.getLogger("pygfx")
warned_for = set()


def resolve_shadercode(wgsl):
    """Apply all shader resolve opearions."""
    return _resolve(wgsl, OutputResolver(), VaryingResolver())


def resolve_varyings(wgsl):
    return _resolve(wgsl, VaryingResolver())


def resolve_output(wgsl):
    return _resolve(wgsl, OutputResolver())


def _resolve(wgsl, *resolvers):
    assert isinstance(wgsl, str)

    # Split into lines, which is easier to process. Ensure it ends with newline in the end.
    lines = wgsl.splitlines()
    while lines and not lines[0]:
        lines.pop(0)
    if not (lines and lines[-1] == ""):
        lines.append("")

    curly_braces_level = 0
    current_func_kind = None

    shared_context = {"more_lines": []}
    for resolver in resolvers:
        resolver.context = shared_context

    # Apply resolve functions. They modify the lines
    for linenr, ori_line in enumerate(lines):
        line = ori_line.split("//", 1)[0].strip()
        if not line:
            continue

        if line.startswith("fn "):
            prevline = lines[linenr - 1].strip() if linenr > 0 else ""
            current_func_kind = "normal"
            if prevline.startswith("@"):
                if prevline.startswith("@vertex"):
                    current_func_kind = "vertex"
                elif prevline.startswith("@fragment"):
                    current_func_kind = "fragment"
                elif prevline.startswith("@compute"):
                    current_func_kind = "compute"
            curly_braces_level = 0

        if current_func_kind:
            curly_braces_level += line.count("{") - line.count("}")
            if curly_braces_level <= 0:
                current_func_kind = None

        for resolver in resolvers:
            resolver.process_line(linenr, line, current_func_kind)

    for resolver in resolvers:
        resolver.update_lines(lines)

    # Return as str
    return "\n".join(lines)


def indent_from_line(line):
    return line[: len(line) - len(line.lstrip())]


varying_types = ["f32", "vec2<f32>", "vec3<f32>", "vec4<f32>"]
varying_types = (
    varying_types
    + [t.replace("f", "i") for t in varying_types]
    + [t.replace("f", "u") for t in varying_types]
)

re_varying_getter = re.compile(r"[^\w]varyings\.(\w+)", re.UNICODE)
re_varying_setter = re.compile(r"varyings\.(\w+)(\.\w+)?\s*\=", re.UNICODE)
builtin_varyings = {"position": "vec4<f32>"}


class VaryingResolver:
    """Resolve varyings in the given wgsl:
    * Detect varyings being used.
    * Check that these are also set.
    * Remove assignments of varyings that are not used.
    * Include the Varyings struct.
    """

    def __init__(self):
        # Prepare dicts that map name to list-of-linenr. And a tupe dict.
        self.assigned_varyings = {}
        self.used_varyings = {}
        self.types = {}  # varying types
        self.context = {"more_lines": []}

        # We try to find the function that first uses the Varyings struct.
        self.struct_linenr = None

        # Go over all lines to:
        # - find the lines where a varying is set
        # - collect the types of these varyings
        # - collect all used varyings
        # - find where the vertex-shader starts

    def process_line(self, linenr, line, current_func_kind):
        # Look for varyings being set (in the vertex shader)
        # Note that this implicitly matches from the start of the stripped line.
        in_vertex_shader = current_func_kind == "vertex"

        # Put the Varyings struct right above where its first used
        if self.struct_linenr is None:
            if "Varyings" in line:
                self.struct_linenr = linenr if current_func_kind else 0

        # Look for varyings being set (in vertex shader)
        if in_vertex_shader:
            match = re_varying_setter.match(line)
            if match:
                # Get parts
                name = match.group(1)
                attr = match.group(2)
                # Handle builtin
                if name in builtin_varyings:
                    self.used_varyings[name] = []
                    self.types[name] = builtin_varyings[name]
                # Find type
                type = line[match.end() :].split("(")[0].strip().replace(" ", "")
                if type not in varying_types:
                    type = ""
                # Triage
                if attr:
                    pass  # Not actually a type but an attribute access
                elif not type:
                    raise TypeError(
                        f"Varying {name!r} assignment needs an explicit cast (of a correct type), e.g. `varying.{name} = f32(3.0);`:\n{line}"
                    )
                elif name in self.types and type != self.types[name]:
                    raise TypeError(
                        f"Varying {name!r} assignment does not match expected type {self.types[name]}:\n{line}"
                    )
                else:
                    self.types[name] = type
                # Store position
                self.assigned_varyings.setdefault(name, []).append(linenr)

        # Look for varyings being get (in any function)
        if current_func_kind and "varyings." in line:
            for match in re_varying_getter.finditer(" " + line):
                name = match.group(1)
                this_varying_is_set_on_this_line = linenr in self.assigned_varyings.get(
                    name, []
                )
                if this_varying_is_set_on_this_line:
                    pass
                elif in_vertex_shader:
                    # If varyings are used in another way than setting, in the vertex shader,
                    # we should either consider them "used", or possibly break the shader if
                    # the used varying is disabled. So let's just not allow it.
                    raise TypeError(
                        f"Varying {name!r} is read in the vertex shader, but only writing is allowed:\n{line}"
                    )
                else:
                    self.used_varyings.setdefault(name, []).append(linenr)

        # For a few names we always declare them as assigned, so that when they're used, but not *really* assigned, they're all zeros, which can be handled.
        # In practice, it means that for (custom) shaders that don't set these varyings, stochastic alpha_method still works, although maybe a bit less pretty.
        for name, type in [("elementIndex", "u32")]:
            self.assigned_varyings.setdefault(name, [])
            self.types[name] = type

    def update_lines(self, lines):
        assigned_varyings = self.assigned_varyings
        used_varyings = self.used_varyings
        types = self.types
        struct_linenr = self.struct_linenr

        # Collect use of varyings in output virtualfields
        for line in self.context["more_lines"]:
            if "varyings." in line:
                for match in re_varying_getter.finditer(" " + line):
                    name = match.group(1)
                    self.used_varyings.setdefault(name, [])

        # Check if all used varyings are assigned
        for name in used_varyings:
            if name not in assigned_varyings:
                try:
                    line = lines[used_varyings[name][0]]
                except Exception:
                    line = ""
                raise TypeError(f"Varying {name!r} is read, but not assigned:\n{line}")

        # Comment-out the varying setter if its unused elsewhere in the shader
        for name, linenrs in assigned_varyings.items():
            if name not in used_varyings:
                for linenr in linenrs:
                    line = lines[linenr]
                    indent = indent_from_line(line)
                    lines[linenr] = indent + "// unused: " + line[len(indent) :]
                    # Deal with multiple-line assignments
                    line_s = line.split("//")[0].strip()
                    while not line_s.endswith(";"):
                        linenr += 1
                        line = lines[linenr].lstrip()
                        line_s = line.split("//")[0].strip()
                        unexpected = "fn ", "struct ", "var ", "let ", "}"
                        if line.startswith(unexpected) or linenr == len(lines) - 1:
                            raise TypeError(
                                f"Varying {name!r} assignment seems to be missing a semicolon:\n{line}"
                            )
                        lines[linenr] = indent + "// " + line

        # Build and insert the struct
        if struct_linenr is not None:
            # Refine position
            while struct_linenr > 0 and not lines[struct_linenr].lstrip().startswith(
                "fn"
            ):
                struct_linenr -= 1
            if struct_linenr > 0 and lines[struct_linenr - 1].lstrip().startswith("@"):
                struct_linenr -= 1
            # First divide into slot-based and builtins
            used_varyings = set(used_varyings)
            used_builtins = used_varyings.intersection(builtin_varyings)
            used_slots = used_varyings.difference(used_builtins)
            used_slots = list(sorted(used_slots))
            # Build struct
            struct_lines = ["struct Varyings {"]
            for slotnr, name in enumerate(used_slots):
                struct_lines.append(f"    @location({slotnr}) {name} : {types[name]},")
            for name in sorted(used_builtins):
                struct_lines.append(f"    @builtin({name}) {name} : {types[name]},")
            struct_lines.append("};\n")
            # Apply indentation and insert
            line = lines[struct_linenr]
            indent = indent_from_line(line)
            new_lines = [indent + line for line in struct_lines] + [line]
            lines[struct_linenr] = "\n".join(new_lines)
        else:
            assert not used_varyings, "woops, did not expect used_varyings here"


re_out_create = re.compile(r"var\s*out\s*\:\s*FragmentOutput\s*;")
re_out_create_old = re.compile(r"var\s*out\s*\=\s*get_fragment_output\(")
re_out_field = re.compile(r"out\.(\w+)\s*\=", re.UNICODE)
re_out_return = re.compile(r"return\s*out\s*;", re.MULTILINE)


class OutputResolver:
    """When out.depth is set (in the fragment shader), adjust the FragmentOutput
    to accept depth.
    """

    def __init__(self):
        # Detect whether the depth is set in the shader. We're going to assume
        # this is in the fragment shader. We check for "out.depth =".
        # Background: by default the depth is based on the geometry (set
        # by vertex shader and interpolated). It is possible for a fragment
        # shader to write the depth instead. If this is done, the GPU cannot
        # do early depth testing; the fragment shader must be run for the
        # depth to be known.

        self.struct_linenr = None
        self.assigned_fields = None  # list of (field-name, linenr)
        self.assigned_fields_list = []  # list of assigned_fields
        self.context = {"more_lines": []}

    def process_line(self, linenr, line, current_func_kind):
        in_fragment_shader = current_func_kind == "fragment"

        if not in_fragment_shader:
            self.assigned_fields = None

        if self.struct_linenr is None:
            if line.startswith("struct FragmentOutput {"):
                self.struct_linenr = linenr

        if in_fragment_shader:
            if self.assigned_fields is None:
                # Detect var out: FragmentOutput
                if re_out_create.match(line) or re_out_create_old.match(line):
                    self.assigned_fields = [("__create", linenr)]
                    self.assigned_fields_list.append(self.assigned_fields)
            else:
                # Detect fields being set
                match = re_out_field.match(line)
                if match:
                    name = match.group(1)
                    self.assigned_fields.append((name, linenr))
                # Detect return, because that's the best place to insert our call
                match = re_out_return.match(line)
                if match:
                    self.assigned_fields.append(("return", linenr))
                    self.assigned_fields = None  # note that this list is an element in assigned_fields_list

    def update_lines(self, lines):
        struct_linenr = self.struct_linenr
        assigned_fields_list = self.assigned_fields_list

        # Collect virtual fields, specified in comments inside the FragmentOutput definition.
        # These will be the args for apply_virtual_fields_of_fragment_output()
        # E.g. // virtualfield foo : f32 = vec4<f32>(0.0)
        virtual_field_prefix = "// virtualfield "
        virtual_fields = []
        if struct_linenr is not None:
            linenr = struct_linenr
            while linenr < len(lines) - 1:
                linenr += 1
                line = lines[linenr].lstrip()
                if line.startswith(virtual_field_prefix):
                    name_type_value = line[len(virtual_field_prefix) :].split("//")[0]
                    name, colon, type_value = name_type_value.partition(":")
                    type, eq, value = type_value.partition("=")
                    assert colon and eq, (
                        f"A virtualfield needs to be of the form 'name : type = defaultvalue'. {name!r} does not"
                    )
                    virtual_fields.append(
                        (name.strip(), type.strip(), value.strip(" ;"))
                    )
                else:
                    break

        # Allow next resolvers to analyze our added lines
        self.context["more_lines"] += [
            virtual_field[2] for virtual_field in virtual_fields
        ]

        depth_is_set = False

        # Handle each place where a FragmentOutput is returned
        for assigned_fields in assigned_fields_list:
            create_linenr = next(i for name, i in assigned_fields if name == "__create")
            create_indent = indent_from_line(lines[create_linenr])

            # Handle get_fragment_output deprecated usage. Not pretty, but it works and is temporary.
            # TODO: after one or two releases, remove the deprecated get_fragment_output case.
            create_line = lines[create_linenr]
            if re_out_create_old.match(create_line.lstrip()):
                if "get_fragment_output" not in warned_for:
                    warned_for.add("get_fragment_output")
                    logger.warning("get_fragment_output() in WGSL is deprecated")
                new_create_line = create_indent + "var out: FragmentOutput;"
                colorset_line = (
                    create_indent
                    + "out.color ="
                    + create_line.split("=", 1)[1].replace(");", ").color;")
                )
                lines[create_linenr - 1] += "\n" + new_create_line
                lines[create_linenr] = colorset_line
                assigned_fields.append(("color", create_linenr))
                create_linenr -= 1

            # Collect virtual argument values
            args = []
            for name, type, default_value in virtual_fields:
                linenrs = [i for n, i in assigned_fields if n == name]
                if not linenrs:
                    args.append(default_value)
                else:
                    prefixed_name = f"out_virtualfield_{name}"
                    args.append(prefixed_name)
                    lines[create_linenr] += (
                        f"\n{create_indent}var {prefixed_name}: {type};"
                    )

                    for linenr in linenrs:
                        lines[linenr] = lines[linenr].replace(
                            f"out.{name}", f"{prefixed_name}", 1
                        )

            # Process special arguments
            if any(i for name, i in assigned_fields if name == "depth"):
                depth_is_set = True

            last_linenr = max(i for _, i in assigned_fields)
            last_line = lines[last_linenr]
            last_indent = indent_from_line(last_line)
            is_return_line = last_line.lstrip().startswith("return")
            extra_last_lines = []

            # Apply the virtual fields. The call to the apply-function is
            # inserted right before 'return out', or right after the last struct
            # field is set (if we did not detect a return).
            if args:
                args.insert(0, "&out")
                line = f"{last_indent}apply_virtual_fields_of_fragment_output({', '.join(args)});"
                extra_last_lines.append(line)

            if extra_last_lines:
                last_lines = (
                    [*extra_last_lines, last_line]
                    if is_return_line
                    else [last_line, *extra_last_lines]
                )
                lines[last_linenr] = "\n".join(last_lines)

        # If depth is set, add the depth field to the struct definition.
        if depth_is_set:
            if struct_linenr is None:
                raise TypeError("FragmentOutput definition not found.")
            depth_field = "    @builtin(frag_depth) depth : f32,"
            line = lines[struct_linenr]
            indent = indent_from_line(line)
            new_lines = [line, indent + depth_field]
            lines[struct_linenr] = "\n".join(new_lines)



"""
The six stages of text rendering (also see the README in this subpackage):

* Entrypoint & itemisation
* Font selection
* Shaping
* Glyph generation
* Layout
* Rendering

This namespace contains a function for the three central stages.
Itemisation and layout are implemented in the Text class.
Rendering is implemented in the TextShader.
"""

from ._fontfinder import FontFile  # noqa: F401
from ._fontmanager import FontProps, FontManager, font_manager  # noqa: F401
from ._atlas import glyph_atlas  # noqa: F401
from ._tokenizers import tokenize_text, tokenize_markdown  # noqa: F401
from . import _sdf
from . import _shaper


select_font = font_manager.select_font  # Font selection
shape_text = _shaper.shape_text  # Shaping
generate_glyph = _sdf.generate_glyph  # Glyph generation



import numpy as np
import pylinalg as la
from typing import List
from ._base import WorldObject
from ..utils import array_from_shadertype
from ..utils.enums import BindMode
from ..resources import Buffer
from ._more import Mesh


class Bone(WorldObject):
    """A bone.

    A bone is a part of a skeleton.The skeleton in turn is used by the SkinnedMesh.
    Bones are almost identical to a blank WorldObject.

    """

    def __init__(self, name=""):
        super().__init__(name=name)


class Skeleton:
    """A skeleton.

    A skeleton is a collection of bones that are used to animate a SkinnedMesh.

    Parameters
    ----------
    bones : List[Bone]
        The list of bones.
        Note that a copy will be made, so you can modify the original list without effecting this one.
    bone_inverses : List[np.ndarray] | None
        A list of matrix4x4 that represent the inverse of the world matrix of the individual bones.
        If not provided, they will be auto calculated.
    """

    def __init__(self, bones: List[Bone], bone_inverses=None):
        super().__init__()
        if bones is None:
            bones = []
        self._bones = bones[:]  # Copy the list to avoid external modifications
        if bone_inverses is None:
            bone_inverses = []
        self._bone_inverses = bone_inverses

        count = len(self.bones)
        self.bone_matrices_buffer = Buffer(
            array_from_shadertype(
                {
                    "bone_matrices": "4x4xf4",
                },
                count,
            ),
            force_contiguous=True,
        )

        if len(self.bone_inverses) == 0:
            self.calculate_inverses()

    @property
    def bones(self):
        """
        The array of bones.
        """
        return self._bones

    @property
    def bone_inverses(self):
        """
        An array of matrix4x4 that represent the inverse of the world matrix of the individual bones.
        """
        return self._bone_inverses

    def calculate_inverses(self):
        """Generate the bone_inverses array if not provided in the constructor."""
        self.bone_inverses.clear()
        for bone in self.bones:
            self.bone_inverses.append(bone.world.inverse_matrix)

    def pose(self):
        """Reset the skeleton to the binding-time pose."""
        for i, bone in enumerate(self.bones):
            bone.world.matrix = la.mat_inverse(self.bone_inverses[i])

        for bone in self.bones:
            if bone.parent and isinstance(bone.parent, Bone):
                bone.local.matrix = bone.parent.world.inverse_matrix @ bone.world.matrix
            else:
                bone.local.matrix = bone.world.matrix

            bone.local.position, bone.local.rotation, bone.local.scale = (
                la.mat_decompose(bone.local.matrix)
            )

    def update(self):
        # TODO: we should update the bone matrices buffer automatically by some mechanism.
        # See: https://github.com/pygfx/pygfx/pull/715#issuecomment-2046493145
        """Update the bone matrices buffer."""

        for i, bone in enumerate(self.bones):
            self.bone_matrices_buffer.data[i]["bone_matrices"] = (
                bone.world.matrix @ self.bone_inverses[i]
            ).T

        self.bone_matrices_buffer.update_full()

    def get_bone(self, name):
        for bone in self.bones:
            if bone.name == name:
                return bone
        return None


class SkinnedMesh(Mesh):
    """A skinned mesh.

    A mesh that has a Skeleton with bones that can then be used to animate the vertices of the geometry.

    """

    uniform_type = dict(
        Mesh.uniform_type,
        bind_matrix="4x4xf4",
        bind_matrix_inv="4x4xf4",
    )

    def __init__(self, geometry, material):
        super().__init__(geometry, material)

        self._bind_matrix = np.eye(4, dtype=np.float32)
        self._bind_matrix_inv = np.eye(4, dtype=np.float32)

        self._bind_mode = BindMode.attached

    @property
    def bind_matrix(self):
        """The base matrix that is used for the bound bone transforms."""
        return self._bind_matrix

    @property
    def bind_matrix_inv(self):
        """The base matrix that is used for resetting the bound bone transforms."""
        return self._bind_matrix_inv

    @property
    def bind_mode(self):
        """
        How a skinned mesh is bound to its skeleton. Either "attached" or "detached".

        See :obj:`pygfx.utils.enums.BindMode`:. Default "attached".
        """
        return self._bind_mode

    @bind_mode.setter
    def bind_mode(self, value):
        assert value in BindMode, f"bind_mode must be one of {BindMode}, not {value}"
        self._bind_mode = value

    def _update_world_transform(self):
        # Update when the world transform has changed
        super()._update_world_transform()

        if self.bind_mode == BindMode.attached:
            self._bind_matrix_inv = self.world.inverse_matrix
        elif self.bind_mode == BindMode.detached:
            self._bind_matrix_inv = la.mat_inverse(self.bind_matrix)

        self.uniform_buffer.data["bind_matrix_inv"] = self._bind_matrix_inv.T
        self.uniform_buffer.update_full()

    def bind(self, skeleton: Skeleton, bind_matrix=None):
        """Bind a skeleton to the skinned mesh.
        The bind_matrix gets saved to .bind_matrix property and the .bind_matrix_inv gets calculated.

        Args:
            skeleton (Skeleton): The skeleton created from the bones tree.
            bind_matrix (np.ndarray, optional): The base transform of the skeleton.
        """
        self.skeleton = skeleton

        if bind_matrix is None:
            self.skeleton.calculate_inverses()
            bind_matrix = self.world.matrix

        self._bind_matrix = bind_matrix
        self._bind_matrix_inv = la.mat_inverse(bind_matrix)

        self.uniform_buffer.data["bind_matrix"] = self._bind_matrix.T
        self.uniform_buffer.data["bind_matrix_inv"] = self._bind_matrix_inv.T
        self.uniform_buffer.update_full()

    def pose(self):
        """Reset the skinned mesh to the binding-time pose."""
        self.skeleton.pose()

    def _update_object(self):
        super()._update_object()
        # todo: theoretically, it's not appropriate to update the skeleton here.
        # because skeleton can be shared by multiple skinned meshes, which could result in redundant updates and caculations.
        if self.skeleton:
            self.skeleton.update()



from ._base import Material
from ..resources import Texture, TextureMap
from ..utils import unpack_bitfield, Color, assert_type
from ..utils.enums import (
    EdgeMode,
    ColorMode,
    SizeMode,
    CoordSpace,
    MarkerInt,
    MarkerShape,
    MarkerMode,
    RotationMode,
)


class PointsMaterial(Material):
    """Point default material.

    Renders disks of the given size and color.

    Parameters
    ----------
    size : float
        The size (diameter) of the points in logical pixels. Default 4.
    size_space : str | CoordSpace
        The coordinate space in which the size is expressed ('screen', 'world', 'model'). Default 'screen'.
    size_mode : str | SizeMode
        The mode by which the points are sized. Default 'uniform'.
    color : str | tuple | Color
        The uniform color of the points (used depending on the ``color_mode``).
    color_mode : str | ColorMode
        The mode by which the points are coloured. Default 'auto'.
    edge_mode : str | EdgeMode
        The mode by which the points are edged. Default 'centered'.
    map : TextureMap | Texture
        The texture map specifying the color for each texture coordinate.
    aa : bool
        Whether the points are anti-aliased in the shader. Default False.
    rotation : float
        The rotation of the point marker in radians. Default 0.
    rotation_mode : str | RotationMode
        The mode by which the points are rotated. Default 'uniform'.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class
        <pygfx.Material>`.

    """

    uniform_type = dict(
        Material.uniform_type,
        color="4xf4",
        size="f4",
        rotation="f4",
    )

    def __init__(
        self,
        size=4,
        size_space="screen",
        size_mode="uniform",
        *,
        color=(1, 1, 1, 1),
        color_mode="auto",
        edge_mode="centered",
        map=None,
        aa=False,
        rotation=0,
        rotation_mode="uniform",
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.size = size
        self.size_space = size_space
        self.size_mode = size_mode
        self.color = color
        self.color_mode = color_mode
        self.edge_mode = edge_mode
        self.map = map
        self.aa = aa
        self.rotation = rotation
        self.rotation_mode = rotation_mode

    def _wgpu_get_pick_info(self, pick_value):
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, index=26, x=9, y=9)
        return {
            "vertex_index": values["index"],
            "point_coord": (values["x"] - 256.0, values["y"] - 256.0),
        }

    @property
    def color(self):
        """The color of the points (if map is not set)."""
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_full()

    @property
    def aa(self):
        """Whether the point's visual edge is anti-aliased.

        Aliasing gives prettier results by producing semi-transparent fragments
        at the edges. Lines thinner than one physical pixel are also diminished
        by making them more transparent.

        However, because semi-transparent fragments are introduced, artifacts
        may occur if certain cases. For the same reason, aa only works for the
        "blended" and "weighted" alpha methods.

        Note that by default, pygfx already uses SSAA and/or PPAA to anti-alias
        the total renderered result. Point-based aa is an *additional* visual
        improvement.
        """
        return self._store.aa

    @aa.setter
    def aa(self, aa):
        self._store.aa = bool(aa)

    @property
    def _gfx_effective_aa(self):
        aa_able_methods = ("blended", "weighted")
        return self._store.aa and self.alpha_method in aa_able_methods

    @property
    def color_mode(self):
        """The way that color is applied to the points.

        See :obj:`pygfx.utils.enums.ColorMode`:
        """
        return self._store.color_mode

    @color_mode.setter
    def color_mode(self, value):
        value = value or "auto"
        if value not in ColorMode:
            raise ValueError(
                f"PointsMaterial.color_mode must be a string in {ColorMode}, not {value!r}"
            )
        if value in ["face", "face_map"]:
            raise ValueError(f"PointsMaterial.color_mode does not support {value!r}")
        self._store.color_mode = value

    @property
    def edge_mode(self):
        """The way that edges are applied to the mesh.

        See :obj:`pygfx.utils.enums.EdgeMode`:
        """
        return self._store.edge_mode

    @edge_mode.setter
    def edge_mode(self, value):
        value = value or "centered"
        if value not in EdgeMode:
            raise ValueError(
                f"PointsMaterial.edge_mode must be a string in {EdgeMode}, not {value!r}"
            )
        self._store.edge_mode = value

    @property
    def vertex_colors(self):
        return self.color_mode == ColorMode.vertex

    @vertex_colors.setter
    def vertex_colors(self, value):
        raise DeprecationWarning(
            "vertex_colors is deprecated, use ``color_mode='vertex'``"
        )

    @property
    def size(self):
        """The size (diameter) of the points, in logical pixels (or world/model space if ``size_space`` is set)."""
        return float(self.uniform_buffer.data["size"])

    @size.setter
    def size(self, size):
        self.uniform_buffer.data["size"] = size
        self.uniform_buffer.update_full()

    @property
    def size_space(self):
        """The coordinate space in which the size is expressed.

        See :obj:`pygfx.utils.enums.CoordSpace`:
        """
        return self._store.size_space

    @size_space.setter
    def size_space(self, value):
        value = value or "screen"
        if value not in CoordSpace:
            raise ValueError(
                f"PointsMaterial.size_space must be a string in {CoordSpace}, not {value!r}"
            )
        self._store.size_space = value

    @property
    def size_mode(self):
        """The way that size is applied to the mesh.

        See :obj:`pygfx.utils.enums.SizeMode`:
        """
        return self._store.size_mode

    @size_mode.setter
    def size_mode(self, value):
        value = value or "uniform"
        if value not in SizeMode:
            raise ValueError(
                f"PointsMaterial.size_mode must be a string in {SizeMode}, not {value!r}"
            )
        self._store.size_mode = value

    @property
    def map(self):
        """The texture map specifying the color for each texture coordinate.
        The dimensionality of the map can be 1D, 2D or 3D, but should match the
        number of columns in the geometry's texcoords.
        """
        return self._store.map

    @map.setter
    def map(self, map):
        assert_type("map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.map = map

    @property
    def rotation(self):
        """The rotation of the marker in radians."""
        return float(self.uniform_buffer.data["rotation"])

    @rotation.setter
    def rotation(self, rotation):
        self.uniform_buffer.data["rotation"] = rotation
        self.uniform_buffer.update_full()

    @property
    def rotation_mode(self):
        """The way that rotation is applied to the markers."""
        return self._store.rotation_mode

    @rotation_mode.setter
    def rotation_mode(self, value):
        value = value or "uniform"
        if value not in RotationMode:
            raise ValueError(
                f"PointsMaterial.rotation_mode must be a string in {RotationMode}, not {value!r}"
            )

        self._store.rotation_mode = value

    # todo: sizeAttenuation


class PointsGaussianBlobMaterial(PointsMaterial):
    """A material to render points as Gaussian blobs.

    Renders Gaussian blobs with a standard deviation that is 1/6th of the
    point-size.
    """


class PointsMarkerMaterial(PointsMaterial):
    """A material to render points as markers.

    Markers come in a variety of shapes, and have an edge with a separate color.

    Parameters
    ----------
    marker : str | MarkerShape
        The shape of the marker. Default 'circle'.
    marker_mode : str | MarkerMode
        The mode by which the markers are defined. Default 'uniform'.
    edge_color : str | tuple | Color
        The color of line marking the edge of the markers. Default 'black'.
    edge_width : float
        The width of the edge of the markers. Default 1.
    kwargs : Any
        Additional kwargs will be passed to the :class:`PointsMaterial<pygfx.PointsMaterial>`.

    """

    uniform_type = dict(
        PointsMaterial.uniform_type,
        edge_color="4xf4",
        edge_width="f4",
    )

    def __init__(
        self,
        *,
        marker="circle",
        marker_mode="uniform",
        edge_width=1.0,
        edge_color="black",
        custom_sdf=None,
        edge_color_mode="auto",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.marker = marker
        self.marker_mode = marker_mode
        self.edge_width = edge_width
        self.edge_color = edge_color
        self.edge_color_mode = edge_color_mode
        self.custom_sdf = custom_sdf

    @property
    def edge_color(self):
        """The color of the edge of the markers."""
        return Color(self.uniform_buffer.data["edge_color"])

    @edge_color.setter
    def edge_color(self, edge_color):
        edge_color = Color(edge_color)
        self.uniform_buffer.data["edge_color"] = edge_color
        self.uniform_buffer.update_full()

    @property
    def edge_width(self):
        """The width of the edge of the markers, in logical pixels (or world/model space if ``size_space`` is set)."""
        return float(self.uniform_buffer.data["edge_width"])

    @edge_width.setter
    def edge_width(self, edge_width):
        self.uniform_buffer.data["edge_width"] = float(edge_width)
        self.uniform_buffer.update_full()

    @property
    def marker(self):
        """The type/shape of the markers.

        Supported values:

        * A string from :obj:`pygfx.utils.enums.MarkerShape`.
        * Matplotlib compatible characters: "osD+x^v<>".
        * Unicode symbols: "●○■♦♥♠♣✳▲▼◀▶".
        * Emojis: "❤️♠️♣️♦️💎💍✳️📍".
        * A string containing the value "custom". In this case, the WGSL
          code defined by ``custom_sdf`` will be used.
        """
        # TODO: is marker a good name?
        # Note: MPL calls this 'marker', Plotly calls this 'symbol'
        return self._store.marker

    @marker.setter
    def marker(self, name):
        # Define possible values, see:
        # https://matplotlib.org/stable/api/markers_api.html
        # https://plotly.com/python/marker-style/#custom-marker-symbols

        alt_names = {
            # MPL
            "o": "circle",
            "s": "square",
            "D": "diamond",
            "+": "plus",
            "x": "cross",
            "^": "triangle_up",
            "<": "triangle_left",
            ">": "triangle_right",
            "v": "triangle_down",
            # Unicode
            "●": "circle",
            "○": "ring",
            "■": "square",
            "♦": "diamond",
            "♥": "heart",
            "♠": "spade",
            "♣": "club",
            "✳": "asterix",
            "▲": "triangle_up",
            "▼": "triangle_down",
            "◀": "triangle_left",
            "▶": "triangle_right",
            # Emojis (these may look like their plaintext variants in your editor)
            "❤️": "heart",
            "♠️": "spade",
            "♣️": "club",
            "♦️": "diamond",
            "💎": "diamond",
            "💍": "ring",
            "✳️": "asterix",
            "📍": "pin",
        }

        name = name or "circle"
        if isinstance(name, int):
            try:
                resolved_name = MarkerInt[name]
            except KeyError:
                raise ValueError(
                    f"PointsMarkerMaterial.marker was set using an int ({name}), but that int does not correspond to a marker."
                ) from None
        else:
            resolved_name = alt_names.get(name, name).lower()
            if resolved_name not in MarkerShape:
                raise ValueError(
                    f"PointsMarkerMaterial.marker must be a string in {SizeMode}, or a supported characted, not {name!r}"
                )
        self._store.marker = resolved_name

    @property
    def marker_mode(self):
        """The mode by which markers are defined.

        See :obj:`pygfx.utils.enums.MarkerMode`:

        When using per-vertex markers, the geometry should have a buffer called "markers", containing ints.
        To make marker names to ints, use :obj:`pygfx.utils.enums.MarkerInt`:.
        """
        return self._store.marker_mode

    @marker_mode.setter
    def marker_mode(self, value):
        value = value or "uniform"
        if value not in MarkerMode:
            raise ValueError(
                f"PointsMaterial.marker_mode must be a string in {MarkerMode}, not {value!r}"
            )
        self._store.marker_mode = value

    @property
    def edge_color_mode(self):
        """The way that color is applied to the points.

        Only "uniform", "auto", and "vertex" are supported.

        For `edge_color_mode`, "auto" is an alias for "uniform".

        See :obj:`pygfx.utils.enums.ColorMode`:
        """
        return self._store.edge_color_mode

    @edge_color_mode.setter
    def edge_color_mode(self, value):
        value = value or "auto"
        if value not in ColorMode:
            raise ValueError(
                f"PointsMaterial.edge_color_mode must be a string in {ColorMode}, not {value!r}"
            )
        if value in ["face", "face_map", "vertex_map"]:
            raise ValueError(
                f"PointsMaterial.edge_color_mode does not support {value!r}"
            )
        self._store.edge_color_mode = value

    @property
    def custom_sdf(self):
        """The SDF code for the marker shape when the marker is set to custom.

        Negative values are inside the shape, positive values are outside the
        shape.

        The SDF's takes in two parameters `coords: vec2<f32>` and `size: f32`.
        The first is a WGSL coordinate and `size` is the overall size of
        the texture. The returned value should be the signed distance from
        any edge of the shape. Distances (positive and negative) that are
        less than half the `edge_width` in absolute terms will be colored
        with the `edge_color`. Other negative distances will be colored by
        `color`.
        """
        return self._store.custom_sdf

    @custom_sdf.setter
    def custom_sdf(self, code):
        self._store.custom_sdf = code


class PointsSpriteMaterial(PointsMaterial):
    """A material to render points as sprite images.

    Renders the provided texture at each point position. The images are square
    and sized just like with a PointMaterial. The texture color is multiplied
    with the point's "normal" color (as calculated depending on ``color_mode``).

    The sprite texture is provided via ``.sprite``.
    """

    def __init__(self, *, sprite=None, **kwargs):
        super().__init__(**kwargs)
        self.sprite = sprite

    @property
    def sprite(self):
        """The texture map specifying the sprite image.

        The dimensionality of the map must be 2D. If None, it just shows a
        uniform color.
        """
        return self._store.sprite

    @sprite.setter
    def sprite(self, sprite):
        assert sprite is None or isinstance(sprite, Texture)
        self._store.sprite = sprite


# Idea: PointsSdfMaterial(PointsMaterial) -> a material where the point shape can be defined via an sdf.



import pylinalg as la
from typing import List
from .. import Geometry, Line, LineSegmentMaterial, Bone, WorldObject


class SkeletonHelper(Line):
    """A helper object to assist with visualizing a Skeleton"""

    def __init__(self, wobject: WorldObject, thickness=1.0):
        bones = self._get_bones(wobject)
        positions = []
        colors = []
        if not bones:
            positions.append([0, 0, 0])
            colors.append([0, 0, 0])
        else:
            for bone in bones:
                if bone.parent and isinstance(bone.parent, Bone):
                    positions.append([0, 0, 0])
                    positions.append([0, 0, 0])

                    colors.append([0, 0, 1])
                    colors.append([0, 1, 0])

        super().__init__(
            Geometry(positions=positions, colors=colors),
            LineSegmentMaterial(
                thickness=thickness, color_mode="vertex", depth_test=False
            ),
        )

        self.root = wobject
        self.bones = bones
        self.bone_descendants = [
            b for b in bones if b.parent and isinstance(b.parent, Bone)
        ]

        self.local.matrix = wobject.world.matrix

    def _update_object(self):
        # Update on every draw: the helper matrix always follows the root object
        self.local.matrix = self.root.world.matrix
        super()._update_object()
        self.update()

    def update(self):
        # TODO: we should update it automatically by some mechanism.
        # See: https://github.com/pygfx/pygfx/pull/715#issuecomment-2046493145
        """Update the helper object to match the Skeleton's bones."""
        geometry = self.geometry
        root_matrix_world_inv = self.root.world.inverse_matrix
        positions = geometry.positions
        positions_data = positions.data

        j = 0
        for bone in self.bone_descendants:
            bone_parent = bone.parent
            bone_matrix = root_matrix_world_inv @ bone.world.matrix
            la.mat_decompose_translation(bone_matrix, out=positions_data[j])

            parent_bone_matrix = root_matrix_world_inv @ bone_parent.world.matrix
            la.mat_decompose_translation(parent_bone_matrix, out=positions_data[j + 1])

            j += 2

        positions.update_range()

    def _get_bones(self, obj) -> List[Bone]:
        # Recursively get all bones from the object and its children.
        bones = []
        if isinstance(obj, Bone):
            bones.append(obj)

        for child in obj.children:
            bones.extend(self._get_bones(child))

        return bones



from time import perf_counter_ns

import numpy as np
import pylinalg as la

from ..objects._base import WorldObject
from ..utils.transform import cached


class Camera(WorldObject):
    """Abstract base camera.

    Camera's are world objects and can be placed in the scene, but this is not required.

    The purpose of a camera is to define the viewpoint for rendering a scene.
    This viewpoint consists of its position and orientation (in the world) and
    its projection.

    In other words, it covers the projection of world coordinates to
    normalized device coordinates (NDC), by the (inverse of) the
    camera's own world matrix and the camera's projection transform.
    The former represent the camera's position, the latter is specific
    to the type of camera.
    """

    _FORWARD_IS_MINUS_Z = True

    def __init__(self):
        super().__init__()
        self._last_modified = perf_counter_ns()

        self._view_size = 1.0, 1.0
        self._view_offset = None

    def flag_update(self):
        self._last_modified = perf_counter_ns()

    @property
    def last_modified(self) -> int:
        return max(self._last_modified, self.world.last_modified)

    def set_view_size(self, width, height):
        """Sets the logical size of the target. Set by the renderer; you should typically not use this."""
        self._view_size = float(width), float(height)
        self.flag_update()

    def set_view_offset(
        self,
        full_width: float,
        full_height: float,
        x: float,
        y: float,
        width: float,
        height: float,
    ):
        """Set the offset in a larger viewing frustrum and override the logical size.

        This is useful for advanced use-cases such as multi-window setups or taking tiled screenshots.
        It is the responsibility of the caller to make sure that the ratio of the ``width`` and ``height``
        match that of the canvas/viewport being rendered to, so that the effective ``pixel_ratio`` is isotropic.

        .. code-block:: python

            # Assuming a canvas with a logical size of 640x480 ...

            # Use a custom logical size
            camera.set_view_offset(320, 240, 0, 0, 320, 240)

            # Render the bottom-left corner, sizes in screen-space become larger (relative to the screen)
            camera.set_view_offset(640, 480, 0, 240, 320, 240)

            # Render the bottom-left corner, sizes in screen-space stay the same (relative to the screen)
            camera.set_view_offset(1280, 960, 0, 480, 640, 480)

        Parameters
        ----------
        full_width (float): The full width of the virtual viewing frustrum.
        full_height (float): The full height of the virtual viewing frustrum.
        x (float): The horizontal offset of the curent sub-view.
        y (float): The vertical offset of the curent sub-view.
        width (float): The width of the current sub-view.
        height (float): The height of the current sub-view.

        """
        # Store values
        self._view_offset = vo = {
            "full_width": float(full_width),
            "full_height": float(full_height),
            "x": float(x),
            "y": float(y),
            "width": float(width),
            "height": float(height),
        }
        # Calculate ndc_offset, a value that can be easily applied in the shader using
        # virtual_ndc = ndc.xy * ndc_offset.xy + ndc_offset.zw
        ax = vo["width"] / vo["full_width"]
        ay = vo["height"] / vo["full_height"]
        self._view_offset["ndc_offset"] = (
            ax,
            ay,
            ax + 2.0 * vo["x"] / vo["full_width"] - 1.0,
            -(ay + 2.0 * vo["y"] / vo["full_height"] - 1.0),
        )
        self.flag_update()

    def clear_view_offset(self):
        """Remove the currently set view offset, returning to a normal view."""
        self._view_offset = None
        self.flag_update()

    def _update_projection_matrix(self) -> np.ndarray:
        raise NotImplementedError()

    def get_state(self):
        """Get the state of the camera as a dict."""
        return {}

    def set_state(self, state):
        """Set the state of the camera from a dict."""
        self.flag_update()

    @property
    def view_matrix(self) -> np.ndarray:
        return self.world.inverse_matrix

    @cached
    def projection_matrix(self) -> np.ndarray:
        base = self._update_projection_matrix()
        if self._view_offset is None:
            return base

        view_offset = self._view_offset
        s_x = view_offset["full_width"] / view_offset["width"]
        s_y = view_offset["full_height"] / view_offset["height"]
        d_x = view_offset["x"] / view_offset["full_width"]
        d_y = view_offset["y"] / view_offset["full_height"]
        t_x = +(s_x - 1.0 - 2.0 * s_x * d_x)
        t_y = -(s_y - 1.0 - 2.0 * s_y * d_y)
        ndc_matrix = np.array(
            [
                [s_x, 0.0, 0.0, t_x],
                [0.0, s_y, 0.0, t_y],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            np.float32,
        )
        proj_matrix = ndc_matrix @ base
        proj_matrix.flags.writeable = False
        return proj_matrix

    @cached
    def projection_matrix_inverse(self) -> np.ndarray:
        proj_inv_matrix = la.mat_inverse(self.projection_matrix)
        proj_inv_matrix.flags.writeable = False
        return proj_inv_matrix

    @cached
    def camera_matrix(self) -> np.ndarray:
        cam_matrix = self.projection_matrix @ self.view_matrix
        cam_matrix.flags.writeable = False
        return cam_matrix


class NDCCamera(Camera):
    """A Camera operating in NDC coordinates.

    Its projection matrix is the identity transform (but its position and rotation can still be set).

    In the NDC coordinate system of wgpu (and Pygfx), x and y are in
    the range -1..1, z is in the range 0..1, and (-1, -1, 0) represents
    the bottom left corner.
    """

    def __init__(self):
        super().__init__()
        self._ndc_proj_matrix = np.eye(4, dtype=float)
        self._ndc_proj_matrix.flags.writeable = False

    def _update_projection_matrix(self):
        return self._ndc_proj_matrix


class ScreenCoordsCamera(Camera):
    """A Camera operating in screen coordinates.

    The depth range is the same as in NDC (0 to 1).
    """

    def __init__(self, invert_y=False):
        super().__init__()
        self._invert_y = bool(invert_y)

    def _update_projection_matrix(self):
        width, height = self._view_size
        sx, sy, sz = 2 / width, 2 / height, 1
        dx, dy, dz = -1, -1, 0
        if self._invert_y:
            dy = -dy
            sy = -sy
        m = sx, 0, 0, dx, 0, sy, 0, dy, 0, 0, sz, dz, 0, 0, 0, 1
        proj_matrix = np.array(m, dtype=float).reshape(4, 4)
        proj_matrix.flags.writeable = False
        return proj_matrix



import numpy as np
import pylinalg as la

from .. import (
    cone_geometry,
    Geometry,
    Line,
    LineSegmentMaterial,
    Mesh,
    MeshBasicMaterial,
)
from ..utils import Color


class AxesHelper(Line):
    """A WorldObject to indicate the scene's axes.

    Generates three arrows starting at the local origin and pointing into the
    direction of the local x, y, and z-axis respectively. Each arrow is colored
    to represent the respective axis. In particular, the x-axis arrow is blue,
    the y-axis arrow is green, and the z-axis arrow is red.

    Parameters
    ----------
    size : float
        The length of the lines in local space.
    thickness : float
        The thickness of the lines in (onscreen) pixels.

    """

    def __init__(self, size=1.0, thickness=2):
        line_positions = np.array(
            [
                [0, 0, 0],
                [1, 0, 0],
                [0, 0, 0],
                [0, 1, 0],
                [0, 0, 0],
                [0, 0, 1],
            ],
            dtype="f4",
        )

        colors = np.array(
            [
                [1, 0, 0, 1],
                [1, 0, 0, 1],  # x is red
                [0, 1, 0, 1],
                [0, 1, 0, 1],  # y is green
                [0, 0, 1, 1],
                [0, 0, 1, 1],  # z is blue
            ],
            dtype="f4",
        )

        arrow_radius = size * 0.1
        # the radius of the cone is the thickness, so that the arrow is twice as wide
        # as the line it sits on.
        # we want the arrow head to maintain the proportions of a equilateral triangle
        # when viewed from the side, so the desired height can be computed
        # by multiplying the radius by sqrt(3)
        arrow_size = np.sqrt(3) * arrow_radius
        cone = cone_geometry(radius=arrow_radius, height=arrow_size)

        line_size = np.max([0.1, size - arrow_size])  # ensure > 0.0
        line_positions *= line_size

        geometry = Geometry(positions=line_positions, colors=colors)
        material = LineSegmentMaterial(
            color_mode="vertex", thickness=thickness, aa=True
        )

        super().__init__(geometry, material)

        for pos, color in zip(line_positions[1::2], colors[1::2], strict=True):
            material = MeshBasicMaterial(color=color)
            arrow_head = Mesh(cone, material)
            arrow_head.local.position = pos
            # offset by half of height since the cones
            # are centered around the origin
            # Do not use +=
            # see: https://github.com/pygfx/pygfx/issues/651
            arrow_head.local.position = (
                arrow_head.local.position + arrow_size / 2 * la.vec_normalize(pos)
            )
            arrow_head.local.rotation = la.quat_from_vecs(
                (0, 0, 1), la.vec_normalize(pos)
            )
            self.add(arrow_head)

    def set_colors(self, x, y, z):
        """Update arrow colors.

        Parameters
        ----------
        x : int, float, str, tuple
            The color of the x arrow. This is a either a single int or float
            (gray), a 4-tuple ``(r,g,b,a)`` of ints or floats, or a hex-coded
            color string in one of the following formats: ``#RGB``, ``#RGBA``,
            ``#RRGGBB``, ``#RRGGBBAA``.
        y : int, float, str, tuple
            The color of the x arrow. This is a either a single int or float
            (gray), a 4-tuple ``(r,g,b,a)`` of ints or floats, or a hex-coded
            color string in one of the following formats: ``#RGB``, ``#RGBA``,
            ``#RRGGBB``, ``#RRGGBBAA``.
        z : int, float, str, tuple
            The color of the x arrow. This is a either a single int or float
            (gray), a 4-tuple ``(r,g,b,a)`` of ints or floats, or a hex-coded
            color string in one of the following formats: ``#RGB``, ``#RGBA``,
            ``#RRGGBB``, ``#RRGGBBAA``.

        """

        x, y, z = Color(x), Color(y), Color(z)
        # update lines
        self._geometry.colors.data[0] = x
        self._geometry.colors.data[1] = x
        self._geometry.colors.data[2] = y
        self._geometry.colors.data[3] = y
        self._geometry.colors.data[4] = z
        self._geometry.colors.data[5] = z
        self._geometry.colors.update_full()
        # update arrow heads
        for arrow, color in zip(self.children, [x, y, z], strict=True):
            arrow.material.color = color



"""
A transform gizmo to manipulate world objects.
"""

import numpy as np
import pylinalg as la

from ..objects import Line, Mesh
from ..geometries import Geometry, sphere_geometry, cone_geometry, box_geometry
from ..materials import MeshBasicMaterial, LineMaterial
from ..objects import WorldObject
from ..utils.viewport import Viewport
from ..utils.transform import AffineTransform


# Colors in hsluv space - https://www.hsluv.org/
# With H: 0/120/240, S: 100, L: 50
WHITE = "#ffffff"
RED = "#ea0064"
GREEN = "#3f8700"
BLUE = "#007eb7"

THICKNESS = 3

NoneType = type(None)


class TransformGizmo(WorldObject):
    """Gizmo to manipulate a WorldObject.

    This Gizmo allows to interactively control WorldObjects with the mouse
    inside a canvas. It can translate and rotate objects relative to the world
    frame, local frame, or camera frame (screen space).

    To control the Gizmo:

    * Click the center sphere to toggle between object-space, world-space and
      screen-space.
    * Grab the center sphere for uniform scaling.
    * Grab the cubes for one-directional scaling (only in object-space).
    * Grab the arrows to translate in one direction.
    * Grab the planes to translate in two directions.
    * Grab the spheres to rotate.

    Parameters
    ----------
    object : WorldObject
        The controlled world object.
    screen_size : float
        The approximate size of the widget in logical pixels. Default 100.

    """

    def __init__(self, object=None, screen_size=100):
        super().__init__()
        self._object_to_control = None

        # We store these as soon as we get a call in ``add_default_event_handlers``
        self._viewport = None
        self._camera = None
        self._ndc_to_screen = None

        # A dict that stores the state at the start of a drag. Or None.
        self._ref = None
        self.gizmo_scale = np.ones(3, dtype=float)

        # The radius (in pixels) the gizmo's screen-space bounding box should occupy
        # (aka. the desired on-screen size of the gizmo)
        self._screen_size = float(screen_size)

        # the extent of the gizmo measured along each cardinal direction
        # expressed in the local frame
        # order: right, up, forward, left, down, backward (first 3-tuple is
        # along positive (x, y, z), second 3-tuple is along negative (x, y, z))
        self._local_extents = None

        # Init
        self._create_elements()
        self.toggle_mode("object")  # object, world, screen
        self._highlight()
        self.set_object(object)

    def set_object(self, object: WorldObject):
        """Update the controlled object.

        Parameters
        ----------
        object : WorldObject
            The new controlled object.

        """

        if not isinstance(object, (WorldObject, NoneType)):
            raise ValueError("The object must be None or a WorldObject instance.")

        self._object_to_control = object

    def toggle_mode(self, mode=None):
        """Switch the reference frame.

        Parameters
        ----------
        mode : str
            The reference frame to switch to. Must be one of  ``"object"``
            (local frame), ``"world"`` (world frame), or ``"screen"`` (camera
            frame). If None the next mode (following this order) is selected.

        """

        modes = "object", "world", "screen"
        if not mode:
            mode = {"object": "world", "world": "screen"}.get(self._mode, "object")
        if mode not in modes:
            raise ValueError(f"Invalid mode '{mode}', must be one of {modes}.")
        self._mode = mode
        self._on_mode_switch()  # The gizmo looks a bit different in each mode

    def _create_elements(self):
        """Create lines, handles, etc."""

        # The location of the scale handles, and where the arcs meet the lines.
        self._halfway = halfway = 0.65

        # --- the parts that are at the center

        # Create screen translate handle
        sphere_geo = sphere_geometry(0.07)
        scale_uniform = Mesh(
            sphere_geo,
            MeshBasicMaterial(color=WHITE, pick_write=True),
        )
        scale_uniform.local.scale = (1.5, 1.5, 1.5)

        # --- the parts that are fully in one dimension

        # Create lines
        line_geo = Geometry(positions=[(0, 0, 0), (halfway, 0, 0)])
        line_x = Line(
            line_geo,
            LineMaterial(thickness=THICKNESS, color=RED),
        )
        line_y = Line(
            line_geo,
            LineMaterial(thickness=THICKNESS, color=GREEN),
        )
        line_z = Line(
            line_geo,
            LineMaterial(thickness=THICKNESS, color=BLUE),
        )

        # Create 1D translate handles
        cone_geo = cone_geometry(0.1, 0.17)
        cone_geo.positions.data[:] = cone_geo.positions.data[:, ::-1]  # xyz->zyx
        translate_x = Mesh(
            cone_geo,
            MeshBasicMaterial(color=RED, pick_write=True),
        )
        translate_y = Mesh(
            cone_geo,
            MeshBasicMaterial(color=GREEN, pick_write=True),
        )
        translate_z = Mesh(
            cone_geo,
            MeshBasicMaterial(color=BLUE, pick_write=True),
        )
        translate_x.local.position = (1, 0, 0)
        translate_y.local.position = (0, 1, 0)
        translate_z.local.position = (0, 0, 1)

        # Create scale handles
        cube_geo = box_geometry(0.1, 0.1, 0.1)
        scale_x = Mesh(
            cube_geo,
            MeshBasicMaterial(color=RED, pick_write=True),
        )
        scale_y = Mesh(
            cube_geo,
            MeshBasicMaterial(color=GREEN, pick_write=True),
        )
        scale_z = Mesh(
            cube_geo,
            MeshBasicMaterial(color=BLUE, pick_write=True),
        )
        scale_x.local.position = (halfway, 0, 0)
        scale_y.local.position = (0, halfway, 0)
        scale_z.local.position = (0, 0, halfway)

        # --- the parts that are in a plane

        # Create arcs
        t = np.linspace(0, np.pi / 2, 64)
        arc_positions = np.stack([0 * t, np.sin(t), np.cos(t)], 1).astype(np.float32)
        arc_geo = Geometry(positions=arc_positions * halfway)
        arc_yz = Line(
            arc_geo,
            LineMaterial(thickness=THICKNESS, color=RED),
        )
        arc_zx = Line(
            arc_geo,
            LineMaterial(thickness=THICKNESS, color=GREEN),
        )
        arc_xy = Line(
            arc_geo,
            LineMaterial(thickness=THICKNESS, color=BLUE),
        )

        # Create in-plane translate handles
        plane_geo = box_geometry(0.01, 0.15, 0.15)
        translate_yz = Mesh(
            plane_geo,
            MeshBasicMaterial(color=RED, pick_write=True),
        )
        translate_zx = Mesh(
            plane_geo,
            MeshBasicMaterial(color=GREEN, pick_write=True),
        )
        translate_xy = Mesh(
            plane_geo,
            MeshBasicMaterial(color=BLUE, pick_write=True),
        )
        inside_arc = 0.4 * halfway
        translate_yz.local.position = (0, inside_arc, inside_arc)
        translate_zx.local.position = (inside_arc, 0, inside_arc)
        translate_xy.local.position = (inside_arc, inside_arc, 0)

        # Create rotation handles
        # These are positioned on each mode switch
        rotate_yz = Mesh(
            sphere_geo,
            MeshBasicMaterial(color=RED, pick_write=True),
        )
        rotate_zx = Mesh(
            sphere_geo,
            MeshBasicMaterial(color=GREEN, pick_write=True),
        )
        rotate_xy = Mesh(
            sphere_geo,
            MeshBasicMaterial(color=BLUE, pick_write=True),
        )

        # --- post-process

        # Rotate objects to their correct orientation
        ob: WorldObject
        for ob in [line_y, translate_y, scale_y, translate_zx, rotate_zx]:
            ob.local.rotation = la.quat_from_axis_angle((0, 0, 1), np.pi / 2)
        for ob in [line_z, translate_z, scale_z, translate_xy, rotate_xy]:
            ob.local.rotation = la.quat_from_axis_angle((0, -1, 0), np.pi / 2)

        arc_xy.local.rotation = la.quat_from_axis_angle((0, 1, 0), np.pi / 2)
        arc_zx.local.rotation = la.quat_from_axis_angle((0, 0, 1), -np.pi / 2)

        # Store the objects
        self._center_sphere = scale_uniform
        self._line_children = line_x, line_y, line_z
        self._arc_children = arc_yz, arc_zx, arc_xy
        self._translate1_children = translate_x, translate_y, translate_z
        self._translate2_children = translate_yz, translate_zx, translate_xy
        self._translate_children = self._translate1_children + self._translate2_children
        self._scale_children = scale_x, scale_y, scale_z
        self._rotate_children = rotate_yz, rotate_zx, rotate_xy

        # Lines and arcs are never opaque. That way they're also not
        # pickable, so they won't be "in the way".
        for ob in self._line_children + self._arc_children:
            ob.material.opacity = 0.6

        # Assign dimensions
        scale_uniform.dim = None
        for triplet in [
            self._line_children,
            self._translate1_children,
            self._scale_children,
            self._rotate_children,
        ]:
            for i, ob in enumerate(triplet):
                ob.dim = i
        for i, ob in enumerate(self._translate2_children):
            dim = [0, 1, 2]
            dim.pop(i)
            ob.dim = tuple(dim)

        # Attach to the gizmo object
        self.add(*self._line_children)
        self.add(*self._arc_children)
        self.add(self._center_sphere)
        self.add(*self._translate_children)
        self.add(*self._scale_children)
        self.add(*self._rotate_children)

        # work out local extent
        self._local_extents = np.empty((6, 3), dtype=float)
        scales = self.get_bounding_box().ravel()  # we know our bb is not None
        self._local_extents[:3] = np.diag(scales[3:])
        self._local_extents[3:] = np.diag(scales[:3])

    def _on_mode_switch(self):
        """When the mode is changed, some adjustments are made."""

        # Note: the elements are affected by the mode in various ways.
        # Much of that (e.g. visibility) is already handled in the
        # _update_xx functions, so we don't have to do that here.

        # Update position of rotation handles
        rotate_yz, rotate_zx, rotate_xy = self._rotate_children
        halfway = self._halfway
        on_arc = halfway * 2**0.5 / 2
        rotate_yz.local.position = (0, on_arc, on_arc)
        rotate_zx.local.position = (on_arc, 0, on_arc)
        rotate_xy.local.position = (on_arc, on_arc, 0)

    def _highlight(self, object=None):
        """Change the appearance during interaction for visual feedback
        on what is being manipulated.
        """
        # Reset thickness of all lines
        for ob in self._line_children:
            if ob.material.thickness != THICKNESS:
                ob.material.thickness = THICKNESS

        for ob in self._arc_children:
            if ob.material.thickness != THICKNESS:
                ob.material.thickness = THICKNESS

        # Set thickness of lines corresponding to the object
        if object:
            lines = []
            dim = object.dim
            if dim is None:  # center handle
                pass
            elif isinstance(dim, int):
                if object in self._rotate_children:
                    lines.append(self._arc_children[dim])
                else:  # translate1 or scale
                    lines.append(self._line_children[dim])
            else:  # translate2
                lines.append(self._line_children[dim[0]])
                lines.append(self._line_children[dim[1]])
            for ob in lines:
                ob.material.thickness *= 1.75

    def update_gizmo(self, event):
        if event.type != "before_render":
            return

        # Note that we almost always update the transform (scale,
        # rotation, position) which means the matrix changed, and so
        # does the world_matrix of all children. In effect the uniforms
        # of all elements need updating anyway, so any other changes
        # to wobject properties (e.g. visibility) are "free" - no need
        # to only update if it actually changes.
        if not self._object_to_control:
            self.visible = False
        elif self._viewport and self._camera:
            self.visible = True
            self._update_ndc_screen_transform()
            self._update_directions()
            self._update_gizmo_transform()
            self._update_visibility()

    def _update_ndc_screen_transform(self):
        # Note: screen origin is at top left corner of NDC with Y-axis pointing down
        x_dim, y_dim = self._viewport.logical_size
        screen_space = AffineTransform()
        screen_space.position = (-1, 1, 0)
        screen_space.scale = (2 / x_dim, -2 / y_dim, 1)
        self._ndc_to_screen = screen_space.inverse_matrix
        self._screen_to_ndc = screen_space.matrix

    def _update_directions(self):
        """
        Calculate how much 1 unit of translation in the draggable space (aka
        mode) translates the object in world and screen space.

        """

        # work out the transforms between the spaces
        camera = self._camera
        if self._mode == "object":
            # local space is draggable
            local_to_world = self._object_to_control.world.matrix
            local_to_ndc = camera.camera_matrix @ local_to_world
            local_to_screen = self._ndc_to_screen @ local_to_ndc
        elif self._mode == "world":
            # world space is draggable
            local_to_world = np.eye(4)
            local_to_ndc = camera.camera_matrix @ local_to_world
            local_to_screen = self._ndc_to_screen @ local_to_ndc
        elif self._mode == "screen":
            # camera space is draggable
            local_to_world = camera.world.matrix
            local_to_ndc = camera.projection_matrix
            local_to_screen = self._ndc_to_screen @ local_to_ndc
        else:  # This cannot happen, in theory
            raise RuntimeError(f"Unexpected mode: `{self._mode}`")

        # points referring to local coordinate axes and origin
        local_points = np.zeros((4, 3))
        local_points[1:, :] = np.eye(3)
        if self._mode == "screen":
            # reference frame has a z-offset from screen origin
            object_to_ndc = camera.camera_matrix @ self._object_to_control.world.matrix
            depth = la.vec_transform((0, 0, 0), object_to_ndc)[2]

            local_points[3, 2] = -1  # camera has inverted Z axis
            local_points[:, 2] -= depth

        # express unit vectors and origin in the various frames
        world_points = la.vec_transform(local_points, local_to_world)
        ndc_points = la.vec_transform(local_points, local_to_ndc)
        screen_points = la.vec_transform(local_points, local_to_screen)

        # store the directions for future use
        self._world_directions = world_points[1:] - world_points[0]
        self._ndc_directions = ndc_points[1:] - ndc_points[0]
        self._screen_directions = screen_points[1:] - screen_points[0]

    def _update_gizmo_transform(
        self,
    ):
        """

        Set the gizmo's transform to keep it in sync with the object it is
        tracking while accounting for the gizmo's mode.

        Position: Set to match the tracked object.
        Rotation: Set to indicate translation directions of the current mode.
        Scale: Set to have the target on-screen size.

        Note: This function also flips the directions of the gizmo's local axes
        so that handles always point towards the camera.
        """

        self.world.position = self._object_to_control.world.position

        # negative/flipped scale on 2 axes registers as 180° rotation. This will
        # be undone by the rotation update below and will sometimes flip the
        # gizmo during rotation updates. Instead, reset scale and restore the
        # desired scale after the rotation update.
        self.world.scale = 1

        if self._mode == "object":
            self.world.rotation = self._object_to_control.world.rotation
        elif self._mode == "world":
            self.world.rotation = np.array((0, 0, 0, 1))
        else:  # self._mode == "screen"
            self.world.rotation = self._camera.world.rotation

        # During interaction we don't update gzimo scale and axis flip, this way:
        # * During a rotation the gizmo does not flip,
        #   making it easier to see how much was rotated.
        # * During a translation the gizmo keeps its "world size",
        #   so that the perspective helps you see how the gizmo has moved.
        if self._ref:
            self.world.scale = self.gizmo_scale
            return

        # size on screen for scale=1
        local_to_screen = (
            self._ndc_to_screen @ self._camera.camera_matrix @ self.world.matrix
        )
        screen_extents = la.vec_transform(self._local_extents, local_to_screen)
        origin_screen = la.vec_transform((0, 0, 0), local_to_screen)
        screen_directions = screen_extents - origin_screen

        # radius of bounding circle (in screen space) and scaling to set to
        # desired radius (aka _screen_size)
        size_1_radius = np.max(np.linalg.norm(screen_directions[:, :2], axis=-1))
        self.gizmo_scale[:] = self._screen_size / size_1_radius

        # if required, flip axes so that handles always point towards the camera
        eps = 1e-10
        should_flip = screen_directions[:3, 2] > eps
        self.gizmo_scale *= 1 - 2 * should_flip

        self.world.scale = self.gizmo_scale

    def _update_visibility(self):
        """Depending on the mode and the orientation of the camera,
        some elements are made invisible.
        """

        # compute the viewing angle onto the gizmo's coordinate planes
        screen_normal = la.vec_transform((0, 0, -1), self._camera.world.matrix)
        screen_normal = la.vec_normalize(screen_normal)
        plane_normal = (
            la.vec_transform(np.eye(3), self.world.matrix) - self.world.position
        )
        plane_normal = la.vec_normalize(plane_normal)
        cos_angle = np.sum(plane_normal * screen_normal, axis=-1)
        viewing_angle = np.pi / 2 - np.arccos(np.clip(np.abs(cos_angle), 0, 1))

        # compute the size of the gizmo's axes on screen
        gizmo_to_screen = (
            self._ndc_to_screen @ self._camera.camera_matrix @ self.world.matrix
        )
        origin_screen = la.vec_transform((0, 0, 0), gizmo_to_screen)
        axes_screen = la.vec_transform(np.eye(3), gizmo_to_screen) - origin_screen
        ax_size = np.linalg.norm(axes_screen[:, :2], axis=-1)

        # check which handles should be shown
        show_1d_translation = ax_size > 30
        show_1d_scaling = ax_size > 30
        show_2d_translation = viewing_angle > deg_to_rad(15)

        if self._mode == "screen":
            show_1d_translation[2] = False
            show_2d_translation[:] = (False, False, True)

        # Per-dimension scaling is only possible in object-mode
        if self._mode != "object":
            show_1d_scaling[:] = False

        # set the visibility
        # Note: uniform scale and rotations are always visible
        for dim in (0, 1, 2):
            self._translate1_children[dim].visible = show_1d_translation[dim]
            self._scale_children[dim].visible = show_1d_scaling[dim]
            self._translate2_children[dim].visible = show_2d_translation[dim]

    # %% Event handling

    def add_default_event_handlers(self, viewport, camera):
        """Register Gizmo callbacks."""

        # Store objects that we need outside the event handling. In
        # contrast to e.g. a controller, the Gizmo also needs to do
        # some calculation at draw time (or right before a draw, to be
        # precise), and for these calculations it needs the viewport
        # and camera.
        viewport = Viewport.from_viewport_or_renderer(viewport)
        self._viewport = viewport
        self._camera = camera

        self.add_event_handler(
            self.process_event,
            "pointer_down",
            "pointer_move",
            "pointer_up",
            "wheel",
        )

        viewport.renderer.add_event_handler(self.update_gizmo, "before_render")

    def process_event(self, event):
        """Callback to handle gizmo-related events."""

        # No interaction if there is no object to control
        if not self._object_to_control:
            return

        # Triage over event type
        type = event.type

        if type == "pointer_down":
            if event.button != 1 or event.modifiers:
                return
            self._ref = None
            # NOTE: I imagine that if multiple tools are active, they
            # each ask for picking info, causing multiple buffer reads
            # for the same location. However, with the new event system
            # this is probably not a problem, when wobjects receive events.
            ob = event.target
            if ob not in self.children:
                return
            # Depending on the object under the pointer, we scale/translate/rotate
            if ob == self._center_sphere:
                self._handle_start("scale", event, ob)
            elif ob in self._translate_children:
                self._handle_start("translate", event, ob)
            elif ob in self._scale_children:
                self._handle_start("scale", event, ob)
            elif ob in self._rotate_children:
                self._handle_start("rotate", event, ob)
            # Highlight the object
            self._highlight(ob)
            self._viewport.renderer.request_draw()
            self.set_pointer_capture(event.pointer_id, event.root)

        elif type == "pointer_up":
            if not self._ref:
                return
            if self._ref["dim"] is None and self._ref["maxdist"] < 3:
                self.toggle_mode()  # clicked on the center sphere
            self._ref = None
            # De-highlight the object
            self._highlight()
            self._viewport.renderer.request_draw()

        elif type == "pointer_move":
            if not self._ref:
                return
            # Get how far we've moved from starting point - we have a dead zone
            dist = (
                (event.x - self._ref["event_pos"][0]) ** 2
                + (event.y - self._ref["event_pos"][1]) ** 2
            ) ** 0.5
            self._ref["maxdist"] = max(self._ref["maxdist"], dist)
            # Delegate to the correct handler
            if self._ref["maxdist"] < 3:
                pass
            elif self._ref["kind"] == "translate":
                self._handle_translate_move(event)
            elif self._ref["kind"] == "scale":
                self._handle_scale_move(event)
            elif self._ref["kind"] == "rotate":
                self._handle_rotate_move(event)
            # Keep viz up to date
            self._viewport.renderer.request_draw()

    def _handle_start(self, kind, event, ob: WorldObject):
        """Initiate a drag. We create a snapshot of the relevant state at this point."""
        this_pos = self._object_to_control.world.position
        ob_pos = ob.world.position
        self._ref = {
            "kind": kind,
            "event_pos": np.array((event.x, event.y)),
            "dim": ob.dim,
            "maxdist": 0,
            # Transform at time of start
            "pos": self._object_to_control.world.position,
            "scale": self._object_to_control.world.scale,
            "rot": self._object_to_control.world.rotation,
            "world_pos": ob_pos,
            "world_offset": ob_pos - this_pos,
            "ndc_pos": la.vec_transform(ob_pos, self._camera.camera_matrix),
            # Gizmo direction state at start-time of drag
            "flips": np.sign(self.world.scale),
            "world_directions": self._world_directions.copy(),
            "ndc_directions": self._ndc_directions.copy(),
            "screen_directions": self._screen_directions.copy(),
        }

    def _handle_translate_move(self, event):
        """Translate action, either using a translate1 or translate2 handle."""

        world_to_screen = self._ndc_to_screen @ self._camera.camera_matrix
        screen_to_world = la.mat_inverse(world_to_screen)

        if isinstance(self._ref["dim"], int):
            travel_directions = (self._ref["dim"],)
        else:
            travel_directions = self._ref["dim"]

        screen_travel = np.array(
            (
                event.x - self._ref["event_pos"][0],
                event.y - self._ref["event_pos"][1],
            )
        )

        # units dragged along gizmo axes
        screen_directions = self._ref["screen_directions"][travel_directions, :]
        if len(screen_directions) == 1:
            # translate 1D: only count movement along translation axis
            units_traveled = get_scale_factor(screen_directions[..., :2], screen_travel)
        else:
            # translate 2D: change basis from screen to gizmo axes
            screen_to_axes = la.mat_inverse(screen_directions[..., :2].T)
            units_traveled = screen_to_axes @ screen_travel

        # pixel units to world units
        # Note: location of translation matters because perspective cameras have
        # shear, i.e., we need to account for start
        start = la.vec_transform(self._ref["world_pos"], world_to_screen)
        end = start + screen_directions.T @ units_traveled
        end_world = la.vec_transform(end, screen_to_world)
        world_units_traveled = end_world - self._ref["world_pos"]

        self._object_to_control.world.position = self._ref["pos"] + world_units_traveled

    def _handle_scale_move(self, event):
        """Scale action."""
        # Get dimension
        dim = self._ref["dim"]

        # Get how the mouse has moved
        screen_moved = np.array(
            (
                event.x - self._ref["event_pos"][0],
                event.y - self._ref["event_pos"][1],
                0,
            )
        )

        if dim is None:
            # Uniform scale
            ref_vec = np.array((1, -1, 0))
            factor = get_scale_factor(ref_vec, screen_moved)
            scale = 2 ** (factor / 100)
            scale = np.array((scale, scale, scale))
        else:
            # Get how much the mouse has moved in the ref direction
            screen_dir = self._ref["screen_directions"][dim]
            factor = get_scale_factor(screen_dir, screen_moved)

            # we flip gizmo axis to point towards the user. As a result,
            # users expect the direction of positive scaling to flip, too
            is_flipped = self.local.scale[dim] < 0
            factor *= 1 - 2 * (is_flipped)

            npixels = factor * np.linalg.norm(screen_dir)
            # Calculate the relative scale
            scale = [1, 1, 1]
            scale[dim] = 2 ** (npixels / 100)
            scale = np.array(scale)

        # Apply
        self._object_to_control.local.scale = scale * self._ref["scale"]

    def _handle_rotate_move(self, event):
        """Rotate action."""

        local_to_screen = (
            self._ndc_to_screen @ self._camera.camera_matrix @ self.world.matrix
        )
        object_screen = la.vec_transform((0, 0, 0), local_to_screen)[:2]

        # amount of cursor rotation around gizmo origin (CCW is positive)
        start_direction = self._ref["event_pos"] - object_screen
        start_angle = np.arctan2(start_direction[1], start_direction[0])
        current_direction = np.array((event.x, event.y)) - object_screen
        current_angle = np.arctan2(current_direction[1], current_direction[0])
        cursor_rotation = current_angle - start_angle

        # axis around which the object rotates
        dim = self._ref["dim"]
        world_axis = self._ref["world_directions"][dim]

        # the cursor rotation is measured around the camera's forward direction
        # (CCW is positive) we need to mirror it if the rotation axis points
        # points the other way (when CW is positive)
        ndc_axis = self._ref["ndc_directions"][dim]
        is_mirrored = 2 * int(ndc_axis[2] > 0) - 1
        cursor_rotation *= is_mirrored

        initial_rotation = self._ref["rot"]
        rotation = la.quat_from_axis_angle(world_axis, cursor_rotation)
        self._object_to_control.world.rotation = la.quat_mul(rotation, initial_rotation)


def get_scale_factor(vec1, vec2):
    """
    Vector project vec2 onto vec1. Aka, figure out how long vec2
    is when measured along vec1.

    This is used, for example, to work out how many units the cursor has
    traveled along a given direction.
    """

    # Note: implementing it like this saves a couple square-roots from
    # normalizing
    return np.sum(vec2 * vec1, axis=-1) / np.sum(vec1**2, axis=-1)


def deg_to_rad(degrees):
    return degrees / 360 * (2 * np.pi)



"""
Containers for Material data.

.. currentmodule:: pygfx.materials

Materials define how a WorldObject is rendered. Many objects support multiple
different materials, e.g. the materials that can be applied to a Mesh object
mostly determine how the object is affected by lighs. Further, the materials
have properties to influence the rendering, like colors, line thickness,
colormaps, the strength of specular reflections, etc.

.. autosummary::
    :toctree: materials/
    :template: ../_templates/custom_layout.rst

    Material
    material_from_trimesh

    MeshAbstractMaterial
    MeshBasicMaterial
    MeshPhongMaterial
    MeshToonMaterial
    MeshNormalMaterial
    MeshNormalLinesMaterial
    MeshSliceMaterial
    MeshStandardMaterial

    PointsMaterial
    PointsGaussianBlobMaterial
    PointsMarkerMaterial
    PointsSpriteMaterial

    LineMaterial
    LineSegmentMaterial
    LineInfiniteSegmentMaterial
    LineArrowMaterial
    LineThinMaterial
    LineThinSegmentMaterial
    LineDebugMaterial

    ImageBasicMaterial

    VolumeBasicMaterial
    VolumeIsoMaterial
    VolumeSliceMaterial
    VolumeRayMaterial
    VolumeMipMaterial
    VolumeMinipMaterial

    BackgroundMaterial
    BackgroundImageMaterial
    BackgroundSkyboxMaterial

    GridMaterial

    TextMaterial

"""

# ruff: noqa: F401

from ._base import Material
from ._compat import material_from_trimesh, material_from_open3d
from ._mesh import (
    MeshAbstractMaterial,
    MeshBasicMaterial,
    MeshPhongMaterial,
    MeshToonMaterial,
    MeshNormalMaterial,
    MeshNormalLinesMaterial,
    MeshSliceMaterial,
    MeshStandardMaterial,
    MeshPhysicalMaterial,
)
from ._points import (
    PointsMaterial,
    PointsGaussianBlobMaterial,
    PointsMarkerMaterial,
    PointsSpriteMaterial,
)
from ._line import (
    LineMaterial,
    LineSegmentMaterial,
    LineInfiniteSegmentMaterial,
    LineArrowMaterial,
    LineThinMaterial,
    LineThinSegmentMaterial,
    LineDebugMaterial,
)
from ._image import ImageBasicMaterial
from ._volume import (
    VolumeBasicMaterial,
    VolumeIsoMaterial,
    VolumeSliceMaterial,
    VolumeRayMaterial,
    VolumeMipMaterial,
    VolumeMinipMaterial,
)
from ._background import (
    BackgroundMaterial,
    BackgroundImageMaterial,
    BackgroundSkyboxMaterial,
)
from ._grid import GridMaterial
from ._text import TextMaterial


# Define __all__ for e.g. Sphinx
__all__ = [
    cls.__name__
    for cls in globals().values()
    if isinstance(cls, type) and issubclass(cls, Material)
]
__all__.sort()
__all__.remove("Material")
__all__.insert(0, "Material")
__all__.extend(["material_from_trimesh"])
__all__.extend(["material_from_open3d"])



"""Provides utilities to deal with color."""

import ctypes
import colorsys
import hsluv

F4 = ctypes.c_float * 4


def _float_from_css_value(v, i, is_hue=False):
    v = v.strip()
    if not is_hue:
        if v.endswith("%"):
            return float(v[:-1]) / 100
        elif i < 3:
            return float(v) / 255
        else:
            return float(v)
    else:
        # Hue is a special case, it can be in degrees or raw number between 0 and 1
        if i == 0:
            return float(v[:-3]) / 360 if v.endswith("deg") else float(v)
        else:
            return float(v[:-1]) / 100 if v.endswith("%") else float(v)


class Color:
    """A representation of color (in the sRGB colorspace).

    Internally the color is stored using 4 32-bit floats (rgba). It can be
    instantiated in a variety of ways. E.g. by providing the color components as
    values between 0 and 1:

        * `Color(r, g, b, a)` providing rgba values.
        * `Color(r, g, b)` providing rgb, alpha is 1.
        * `Color(gray, a)` grayscale intensity and alpha.
        * `Color(gray)` grayscale intensity.

    The above variations can also be supplied as a single tuple/list, or
    anything that:

        * `Color((r, g, b))`.

    Named colors:

        * `Color("red")` base color names.
        * `Color("cornflowerblue")` CSS color names.
        * `Color("m")` Matlab color chars.

    Hex colors:

        * `Color("#ff0000")` the common hex format.
        * `Color("#ff0000ff")` the hex format that includes alpha.
        * `Color("#ff0")` the short form hex format.
        * `Color("#ff0f")` the short form hex format that includes alpha.

    CSS color functions:

        * `Color("rgb(255, 0, 0)")` or `Color("rgb(100%, 0%, 0%)")`.
        * `Color("rgba(255, 0, 0, 1.0)")` or `Color("rgba(100%, 0%, 0%, 100%)")`.
        * `Color("hsv(0.333, 1, 0.5)")` or `Color("hsv(120deg, 100%, 50%)")`.
        * `Color("hsva(0.333, 1, 0.5, 1.0)")` or `Color("hsva(120deg, 100%, 50%, 100%)")`.
        * `Color("hsl(0.333, 1, 0.5)")` or `Color("hsl(120deg, 100%, 50%)")`.
        * `Color("hsla(0.333, 1, 0.5, 1.0)")` or `Color("hsla(120deg, 100%, 50%, 100%)")`.
        * `Color("hsluv(0.333, 0.5, 0.5)")` or `Color("hsluv(120deg, 50%, 50%)")`.
        * `Color("hsluva(0.333, 0.5, 0.5, 1.0)")` or `Color("hsluva(120deg, 50%, 50%, 100%)")`.

    Parameters
    ----------
    args : tuple, int, str
        The color specification. Check the docstring of this function for
        details on available format options.

    """

    # Internally, the color is a ctypes float array
    __slots__ = ["_val"]

    def __init__(self, *args):
        if len(args) == 1:
            color = args[0]
            if isinstance(color, (int, float)):
                self._set_from_tuple(args)
            elif isinstance(color, str):
                self._set_from_str(color)
            else:
                # Assume it's an iterable,
                # may raise TypeError 'object is not iterable'
                self._set_from_tuple(color)
        else:
            self._set_from_tuple(args)

    def __repr__(self):
        # A precision of 4 decimals, i.e. 10001 possible values for each color.
        # We truncate zeros, but make sure the value does not end with a dot.
        f = lambda v: f"{v:0.4f}".rstrip("0").ljust(3, "0")
        return f"Color({f(self.r)}, {f(self.g)}, {f(self.b)}, {f(self.a)})"

    @property
    def __array_interface__(self):
        # Numpy can wrap our memory in an array without copying
        readonly = True
        ptr = ctypes.addressof(self._val)
        x = dict(version=3, shape=(4,), typestr="<f4", data=(ptr, readonly))
        return x

    def __len__(self):
        return 4

    def __getitem__(self, index):
        return self._val[index]

    def __iter__(self):
        return self.rgba.__iter__()

    def __eq__(self, other):
        if not isinstance(other, Color):
            other = Color(other)
        return all(self._val[i] == other._val[i] for i in range(4))

    def __add__(self, other):
        return Color(
            self.r + other.r,
            self.g + other.g,
            self.b + other.b,
            self.a,
        )

    def __mul__(self, factor):
        if not isinstance(factor, (float, int)):
            raise TypeError("Can only multiple a color with a scalar.")
        return Color(
            self.r * factor,
            self.g * factor,
            self.b * factor,
            self.a,
        )

    def __truediv__(self, factor):
        if not isinstance(factor, (float, int)):
            raise TypeError("Can only multiple a color with a scalar.")
        return self.__mul__(1 / factor)

    def _set_from_rgba(self, r, g, b, a):
        a = max(0.0, min(1.0, float(a)))
        self._val = F4(float(r), float(g), float(b), a)

    def _set_from_tuple(self, color):
        color = tuple(float(c) for c in color)
        if len(color) == 4:
            self._set_from_rgba(*color)
        elif len(color) == 3:
            self._set_from_rgba(*color, 1)
        elif len(color) == 2:
            self._set_from_rgba(color[0], color[0], color[0], color[1])
        elif len(color) == 1:
            self._set_from_rgba(color[0], color[0], color[0], 1)
        else:
            raise ValueError(f"Cannot parse color tuple with {len(color)} values")

    def _set_from_str(self, color):
        color = color.lower()
        if color.startswith("#"):
            # A hex number
            if len(color) == 7:  # #rrggbb
                self._set_from_rgba(
                    int(color[1:3], 16) / 255,
                    int(color[3:5], 16) / 255,
                    int(color[5:7], 16) / 255,
                    1,
                )
            elif len(color) == 4:  # #rgb
                self._set_from_rgba(
                    int(color[1], 16) / 15,
                    int(color[2], 16) / 15,
                    int(color[3], 16) / 15,
                    1,
                )
            elif len(color) == 9:  # #rrggbbaa
                self._set_from_rgba(
                    int(color[1:3], 16) / 255,
                    int(color[3:5], 16) / 255,
                    int(color[5:7], 16) / 255,
                    int(color[7:9], 16) / 255,
                )
            elif len(color) == 5:  # #rgba
                self._set_from_rgba(
                    int(color[1], 16) / 15,
                    int(color[2], 16) / 15,
                    int(color[3], 16) / 15,
                    int(color[4], 16) / 15,
                )
            else:
                raise ValueError(
                    f"Expecting 4, 5, 7, or 9 chars in a hex number, got {len(color)}."
                )
        elif color.startswith(("rgb(", "rgba(")):
            # A CSS color 'function'
            parts = color.split("(")[1].split(")")[0].split(",")
            parts = [_float_from_css_value(p, i) for i, p in enumerate(parts)]
            if len(parts) == 3:
                self._set_from_rgba(parts[0], parts[1], parts[2], 1)
            elif len(parts) == 4:
                self._set_from_rgba(parts[0], parts[1], parts[2], parts[3])
            else:
                raise ValueError(
                    f"CSS color {color.split('(')[0]}(..) must have 3 or 4 elements, not {len(parts)} "
                )
        elif color.startswith(("hsl(", "hsla(", "hsv(", "hsva(", "hsluv(", "hsluva(")):
            parts = color.split("(")[1].split(")")[0].split(",")
            parts = [
                _float_from_css_value(p, i, is_hue=True) for i, p in enumerate(parts)
            ]
            if len(parts) == 3 or len(parts) == 4:
                if color.startswith(("hsl(", "hsla(")):
                    color = Color.from_hsl(*parts)
                elif color.startswith(("hsv(", "hsva(")):
                    color = Color.from_hsv(*parts)
                elif color.startswith(("hsluv(", "hsluva(")):
                    color = Color.from_hsluv(*parts)
                self._set_from_rgba(
                    color._val[0], color._val[1], color._val[2], color._val[3]
                )
            else:
                raise ValueError(
                    f"CSS color {color.split('(')[0]}(..) must have 3 or 4 elements, not {len(parts)} "
                )
        else:
            # Maybe a named color
            try:
                color_int = NAMED_COLORS[color.lower()]
            except KeyError:
                raise ValueError(f"Unknown color: '{color}'") from None
            else:
                self._set_from_str(color_int)

    @property
    def rgba(self):
        """The RGBA tuple (values between 0 and 1)."""
        return self._val[0], self._val[1], self._val[2], self._val[3]

    @property
    def rgb(self):
        """The RGB tuple (values between 0 and 1)."""
        return self._val[0], self._val[1], self._val[2]

    @property
    def r(self):
        """The red value."""
        return self._val[0]

    @property
    def g(self):
        """The green value."""
        return self._val[1]

    @property
    def b(self):
        """The blue value."""
        return self._val[2]

    @property
    def a(self):
        """The alpha (transparency) value, between 0 and 1."""
        return self._val[3]

    @property
    def gray(self):
        """Return the grayscale intensity."""
        # Calculate with the weighted (a.k.a. luminosity) method, same as Matlab
        r, g, b = self.rgb
        return 0.2989 * r + 0.5870 * g + 0.1140 * b

    @property
    def hex(self):
        """The CSS hex string, e.g. "#00ff00". The alpha channel is ignored.
        Values are clipped to 00 an ff.
        """
        c = self.clip()
        r = int(c.r * 255 + 0.5)
        b = int(c.b * 255 + 0.5)
        g = int(c.g * 255 + 0.5)
        i = (r << 16) + (g << 8) + b
        return "#" + hex(i)[2:].rjust(6, "0")

    @property
    def hexa(self):
        """The hex string including alpha, e.g. "#00ff00ff".
        Values are clipped to 00 an ff.
        """
        c = self.clip()
        r = int(c.r * 255 + 0.5)
        b = int(c.b * 255 + 0.5)
        g = int(c.g * 255 + 0.5)
        a = int(c.a * 255 + 0.5)
        i = (r << 24) + (g << 16) + (b << 8) + a
        return "#" + hex(i)[2:].rjust(8, "0")

    @property
    def css(self):
        """The CSS color string, e.g. "rgba(0,255,0,0.5)"."""
        r, g, b, a = self.rgba
        if a == 1:
            return (
                f"rgb({int(255 * r + 0.5)},{int(255 * g + 0.5)},{int(255 * b + 0.5)})"
            )
        else:
            return f"rgba({int(255 * r + 0.5)},{int(255 * g + 0.5)},{int(255 * b + 0.5)},{a:0.3f})"

    def clip(self):
        """Return a new Color with the values clipped between 0 and 1."""
        return Color(max(0.0, min(1.0, x)) for x in self.rgba)

    @classmethod
    def from_physical(cls, r, g, b, a=1):
        """Create a Color object from a color in the physical colorspace.

        With the physical colorspace we mean what is sometimes called
        Linear-sRGB. It has the same gamut as sRGB, but where sRGB is
        linear w.r.t. human perception, Linear-sRGB is linear w.r.t.
        lumen and photon counts. Calculations on colors in pygfx's shaders
        are done in the physical colorspace.
        """
        return Color(_physical2srgb(r), _physical2srgb(g), _physical2srgb(b), a)

    def to_physical(self):
        """Get the color represented in the physical colorspace, as 3 floats."""
        return _srgb2physical(self.r), _srgb2physical(self.g), _srgb2physical(self.b)

    @classmethod
    def from_hsv(cls, hue, saturation, value, alpha=1):
        """Create a Color object from a color in the HSV (a.k.a. HSB) colorspace.

        HSV stands for hue, saturation, value (aka brightness). The hue
        component indicates the color tone. Values go from red (0) to
        green (0.333) to blue (0.666) and back to red (1). The
        satutation indicates vividness, with 0 meaning gray and 1
        meaning the primary color. The value/brightness indicates goes
        from 0 (black) to 1 (white).

        The alpha channel is optional and defaults to 1.
        """
        color = Color(colorsys.hsv_to_rgb(hue, saturation, value))
        color._val[3] = alpha
        return color

    def to_hsv(self):
        """Get the color represented in the HSV colorspace, as 3 floats."""
        return colorsys.rgb_to_hsv(*self.rgb)

    def to_hsva(self):
        """Get the color represented in the HSV colorspace, as 4 floats."""
        h, s, v = colorsys.rgb_to_hsv(*self.rgb)
        return h, s, v, self.a

    @classmethod
    def from_hsl(cls, hue, saturation, lightness, alpha=1):
        """Create a Color object from a color in the HSL colorspace.

        The HSL colorspace is similar to the HSV colorspace, except the
        "value" is replaced with "lightness". This lightness scales the
        color differently, e.g. a lightness of 1 always represents full
        white.

        The alpha channel is optional and defaults to 1.
        """
        color = Color(colorsys.hls_to_rgb(hue, lightness, saturation))
        color._val[3] = alpha
        return color

    def to_hsl(self):
        """Get the color represented in the HSL colorspace, as 3 floats."""
        hue, lightness, saturation = colorsys.rgb_to_hls(*self.rgb)
        return hue, saturation, lightness

    def to_hsla(self):
        """Get the color represented in the HSL colorspace, as 4 floats."""
        hue, lightness, saturation = colorsys.rgb_to_hls(*self.rgb)
        return hue, saturation, lightness, self.a

    @classmethod
    def from_hsluv(cls, hue, saturation, lightness, alpha=1):
        """Create a Color object from a color in the HSLuv colorspace.

        HSLuv is a human-friendly alternative to HSL. The hue component works
        the same as HSL/HSV, going from red (0) through green (0.333) and blue (0.666)
        back to red (1). The saturation ranges from 0 (grayscale) to 1 (pure color).
        The lightness ranges from 0 (black) to 1 (white). Unlike HSL/HSV, HSLuv
        provides perceptually uniform brightness and saturation.

        The alpha channel is optional and defaults to 1.
        """
        h, s, light = 360.0 * hue, 100.0 * saturation, 100.0 * lightness
        color = Color(hsluv.hsluv_to_rgb((h, s, light)))
        color._val[3] = alpha
        return color

    def to_hsluv(self):
        """Get the color represented in the HSLuv colorspace, as 3 floats."""
        h, s, light = hsluv.rgb_to_hsluv(self.rgb)
        return h / 360.0, s / 100.0, light / 100.0

    def to_hsluva(self):
        """Get the color represented in the HSLuv colorspace, as 4 floats."""
        h, s, light = hsluv.rgb_to_hsluv(self.rgb)
        return h / 360.0, s / 100.0, light / 100.0, self.a

    def lerp(self, target, t):
        """Linear interpolate from source color towards target color with factor t in RGBA space.

        Parameters
        ----------
        target : Color
            The target color to interpolate towards
        t : float
            Interpolation factor between 0 and 1

        Returns
        -------
        Color
            The interpolated color
        """
        r = self.r + (target.r - self.r) * t
        g = self.g + (target.g - self.g) * t
        b = self.b + (target.b - self.b) * t
        a = self.a + (target.a - self.a) * t
        return Color(r, g, b, a)

    def lerp_in_hue(self, target, t, colorspace="hsluv"):
        """Linear interpolate from source color towards target color with factor t in specified colorspace.

        Parameters
        ----------
        target : Color
            The target color to interpolate towards
        t : float
            Interpolation factor between 0 and 1
        colorspace : str
            The colorspace to interpolate in. One of "hsl", "hsluv", "hsv". Default is "hsluv"

        Returns
        -------
        Color
            The interpolated color
        """
        if colorspace == "hsl":
            h1, s1, l1 = self.to_hsl()
            h2, s2, l2 = target.to_hsl()
            to_rgba = lambda h, s, light, a: Color.from_hsl(h, s, light, a)
        elif colorspace == "hsluv":
            h1, s1, l1 = self.to_hsluv()
            h2, s2, l2 = target.to_hsluv()
            to_rgba = lambda h, s, light, a: Color.from_hsluv(h, s, light, a)
        elif colorspace == "hsv":
            h1, s1, l1 = self.to_hsv()
            h2, s2, l2 = target.to_hsv()
            to_rgba = lambda h, s, light, a: Color.from_hsv(h, s, light, a)
        else:
            raise ValueError(f"Unknown colorspace {colorspace}")

        # Special case for hue - interpolate along shortest path
        if abs(h2 - h1) > 0.5:
            if h1 > h2:
                h2 += 1.0
            else:
                h1 += 1.0
        h = (h1 + (h2 - h1) * t) % 1.0
        s = s1 + (s2 - s1) * t
        light = l1 + (l2 - l1) * t
        a = self.a + (target.a - self.a) * t

        return to_rgba(h, s, light, a)

    def lighter(self, factor=0.5, colorspace="hsluv"):
        """Make the color lighter by the given factor.

        Parameters
        ----------
        factor : float
            Factor to lighten by, between 0 and 1. Default is 0.5
        colorspace : str
            The colorspace to use, one of "hsl", "hsluv", or "hsv". Default is "hsluv"

        Returns
        -------
        Color
            A lighter version of the color
        """
        if colorspace not in ["hsl", "hsluv", "hsv"]:
            raise ValueError("colorspace must be one of 'hsl', 'hsluv', or 'hsv'")
        if not 0 <= factor <= 1:
            raise ValueError("factor must be between 0 and 1")

        if colorspace == "hsl":
            h, s, light = self.to_hsl()
            light = light + (1 - light) * factor
            return Color.from_hsl(h, s, light, self.a)
        elif colorspace == "hsluv":
            h, s, light = self.to_hsluv()
            light = light + (1 - light) * factor
            return Color.from_hsluv(h, s, light, self.a)
        else:  # hsv
            h, s, v = self.to_hsv()
            v = v + (1 - v) * factor
            return Color.from_hsv(h, s, v, self.a)

    def darker(self, factor=0.5, colorspace="hsluv"):
        """Make the color darker by the given factor.

        Parameters
        ----------
        factor : float
            Factor to darken by, between 0 and 1. Default is 0.5
        colorspace : str
            The colorspace to use, one of "hsl", "hsluv", or "hsv". Default is "hsluv"

        Returns
        -------
        Color
            A darker version of the color
        """
        if colorspace not in ["hsl", "hsluv", "hsv"]:
            raise ValueError("colorspace must be one of 'hsl', 'hsluv', or 'hsv'")
        if not 0 <= factor <= 1:
            raise ValueError("factor must be between 0 and 1")

        if colorspace == "hsl":
            h, s, light = self.to_hsl()
            light = light * (1 - factor)
            return Color.from_hsl(h, s, light, self.a)
        elif colorspace == "hsluv":
            h, s, light = self.to_hsluv()
            light = light * (1 - factor)
            return Color.from_hsluv(h, s, light, self.a)
        else:  # hsv
            h, s, v = self.to_hsv()
            v = v * (1 - factor)
            return Color.from_hsv(h, s, v, self.a)


def _srgb2physical(c):
    # The simplified version has a maximum error less than 1%, but that's still
    # two steps in the range 0..255.
    # return c ** 2.2
    return c / 12.92 if c <= 0.04045 else ((c + 0.055) / 1.055) ** 2.4


def _physical2srgb(c):
    # return c ** (1 / 2.2)
    return c * 12.92 if c <= 0.0031308 else c ** (1 / 2.4) * 1.055 - 0.055


NAMED_COLORS = {
    # CSS Level 1
    "black": "#000000",
    "silver": "#C0C0C0",
    "gray": "#808080",
    "white": "#FFFFFF",
    "maroon": "#800000",
    "red": "#FF0000",
    "purple": "#800080",
    "fuchsia": "#FF00FF",
    "green": "#008000",
    "lime": "#00FF00",
    "olive": "#808000",
    "yellow": "#FFFF00",
    "navy": "#000080",
    "blue": "#0000FF",
    "teal": "#008080",
    "aqua": "#00FFFF",
    # CSS Level 2
    "orange": "#FFA500",
    # CSS Color Module Level 3
    "aliceblue": "#F0F8FF",
    "antiquewhite": "#FAEBD7",
    "aquamarine": "#7FFFD4",
    "azure": "#F0FFFF",
    "beige": "#F5F5DC",
    "bisque": "#FFE4C4",
    "blanchedalmond": "#FFEBCD",
    "blueviolet": "#8A2BE2",
    "brown": "#A52A2A",
    "burlywood": "#DEB887",
    "cadetblue": "#5F9EA0",
    "chartreuse": "#7FFF00",
    "chocolate": "#D2691E",
    "coral": "#FF7F50",
    "cornflowerblue": "#6495ED",
    "cornsilk": "#FFF8DC",
    "crimson": "#DC143C",
    "cyan": "#00FFFF",
    "darkblue": "#00008B",
    "darkcyan": "#008B8B",
    "darkgoldenrod": "#B8860B",
    "darkgray": "#A9A9A9",
    "darkgreen": "#006400",
    "darkgrey": "#A9A9A9",
    "darkkhaki": "#BDB76B",
    "darkmagenta": "#8B008B",
    "darkolivegreen": "#556B2F",
    "darkorange": "#FF8C00",
    "darkorchid": "#9932CC",
    "darkred": "#8B0000",
    "darksalmon": "#E9967A",
    "darkseagreen": "#8FBC8F",
    "darkslateblue": "#483D8B",
    "darkslategray": "#2F4F4F",
    "darkslategrey": "#2F4F4F",
    "darkturquoise": "#00CED1",
    "darkviolet": "#9400D3",
    "deeppink": "#FF1493",
    "deepskyblue": "#00BFFF",
    "dimgray": "#696969",
    "dimgrey": "#696969",
    "dodgerblue": "#1E90FF",
    "firebrick": "#B22222",
    "floralwhite": "#FFFAF0",
    "forestgreen": "#228B22",
    "gainsboro": "#DCDCDC",
    "ghostwhite": "#F8F8FF",
    "gold": "#FFD700",
    "goldenrod": "#DAA520",
    "greenyellow": "#ADFF2F",
    "grey": "#808080",
    "honeydew": "#F0FFF0",
    "hotpink": "#FF69B4",
    "indianred": "#CD5C5C",
    "indigo": "#4B0082",
    "ivory": "#FFFFF0",
    "khaki": "#F0E68C",
    "lavender": "#E6E6FA",
    "lavenderblush": "#FFF0F5",
    "lawngreen": "#7CFC00",
    "lemonchiffon": "#FFFACD",
    "lightblue": "#ADD8E6",
    "lightcoral": "#F08080",
    "lightcyan": "#E0FFFF",
    "lightgoldenrodyellow": "#FAFAD2",
    "lightgray": "#D3D3D3",
    "lightgreen": "#90EE90",
    "lightgrey": "#D3D3D3",
    "lightpink": "#FFB6C1",
    "lightsalmon": "#FFA07A",
    "lightseagreen": "#20B2AA",
    "lightskyblue": "#87CEFA",
    "lightslategray": "#778899",
    "lightslategrey": "#778899",
    "lightsteelblue": "#B0C4DE",
    "lightyellow": "#FFFFE0",
    "limegreen": "#32CD32",
    "linen": "#FAF0E6",
    "magenta": "#FF00FF",
    "mediumaquamarine": "#66CDAA",
    "mediumblue": "#0000CD",
    "mediumorchid": "#BA55D3",
    "mediumpurple": "#9370DB",
    "mediumseagreen": "#3CB371",
    "mediumslateblue": "#7B68EE",
    "mediumspringgreen": "#00FA9A",
    "mediumturquoise": "#48D1CC",
    "mediumvioletred": "#C71585",
    "midnightblue": "#191970",
    "mintcream": "#F5FFFA",
    "mistyrose": "#FFE4E1",
    "moccasin": "#FFE4B5",
    "navajowhite": "#FFDEAD",
    "oldlace": "#FDF5E6",
    "olivedrab": "#6B8E23",
    "orangered": "#FF4500",
    "orchid": "#DA70D6",
    "palegoldenrod": "#EEE8AA",
    "palegreen": "#98FB98",
    "paleturquoise": "#AFEEEE",
    "palevioletred": "#DB7093",
    "papayawhip": "#FFEFD5",
    "peachpuff": "#FFDAB9",
    "peru": "#CD853F",
    "pink": "#FFC0CB",
    "plum": "#DDA0DD",
    "powderblue": "#B0E0E6",
    "rosybrown": "#BC8F8F",
    "royalblue": "#4169E1",
    "saddlebrown": "#8B4513",
    "salmon": "#FA8072",
    "sandybrown": "#F4A460",
    "seagreen": "#2E8B57",
    "seashell": "#FFF5EE",
    "sienna": "#A0522D",
    "skyblue": "#87CEEB",
    "slateblue": "#6A5ACD",
    "slategray": "#708090",
    "slategrey": "#708090",
    "snow": "#FFFAFA",
    "springgreen": "#00FF7F",
    "steelblue": "#4682B4",
    "tan": "#D2B48C",
    "thistle": "#D8BFD8",
    "tomato": "#FF6347",
    "turquoise": "#40E0D0",
    "violet": "#EE82EE",
    "wheat": "#F5DEB3",
    "whitesmoke": "#F5F5F5",
    "yellowgreen": "#9ACD32",
    # CSS Color Module Level 4
    "rebeccapurple": "#663399",
    # Matlab / Matplotlib
    "b": "#0000FF",
    "g": "#00FF00",
    "r": "#FF0000",
    "c": "#00FFFF",
    "m": "#FF00FF",
    "y": "#FFFF00",
    "k": "#000000",
    "w": "#FFFFFF",
}



"""
Text shaping with Harfbuzz and Freetype.

Relevant links:
* https://harfbuzz.github.io/
* https://freetype.org/freetype2/docs/glyphs/glyphs-3.html

"""

import time
import freetype
import uharfbuzz
import numpy as np


# Determine reference size. This affects the size of the SDF bitmap.
REF_GLYPH_SIZE = 48  # 48px == 64pt


def shape_text(text, font_filename, direction=None):
    """Text shaping. Takes a Unicode string, and replace it with a list
    of glyphs, positioned to represent the given text in a good way.
    Depending on the algorithm used, this takes care of kerning,
    ligatures, and much more.

    Parameters:
        text (str): the text to shape.
        font_filename (str): the font to shape for.

    Returns:
        glyph_indices (list): the indices of the glyphs in the font.
        positions (ndarray): the positions at which these glyphs must be placed.
        meta (dict): additional information about this text (width, direction, and more).

    All returned distances are measured in unit font_size.
    """
    return shape_text_hb(text, font_filename, direction)
    # return shape_text_ft(text, font_filename)


class TemporalCache:
    """A simple cache that drops old items based on time."""

    def __init__(self, lifetime, *, getter, minimum_items=0):
        self._ref_lifetime = lifetime
        self._minimum_items = minimum_items
        self._cache = {}
        self._lifetimes = {}
        self._getter = getter
        self._last_key = None
        self._last_val = None

    def __getitem__(self, key):
        """Gets the object corresponding to the given key.

        Will reset the item's lifetime.
        Getting triggers the lifetimes of all items to be checked.
        """
        # Make subsequent calls *really* fast
        if key == self._last_key:
            self._lifetimes[key] = time.time()
            return self._last_val

        # Get result
        try:
            res = self._cache[key]
        except KeyError:
            res = self._getter(key)
            self._cache[key] = res

        self._last_key = key
        self._last_val = res

        # Update cache
        self._lifetimes[key] = time.time()
        if len(self._cache) > self._minimum_items:
            self.check_lifetimes()

        return res

    def __contains__(self, key):
        return key in self._cache

    def __len__(self):
        return len(self._cache)

    def check_lifetimes(self):
        """Check the lifetimes of all objects."""
        # We want this to be fast. I played with only checking a subset,
        # so make it faster if there are say 10k items in the set. This
        # makes it faster for that case, but the added complexity makes
        # it slower already for 100 items, which is way more realistic
        # font count. So let's keep things simple :)
        pretty_old = time.time() - self._ref_lifetime
        old_items = sorted(
            ((key, lt) for key, lt in self._lifetimes.items() if lt < pretty_old),
            key=lambda x: x[1],
        )
        max_items_to_remove = max(len(self._cache) - self._minimum_items, 0)

        for key, _lt in old_items[:max_items_to_remove]:
            self._lifetimes.pop(key, None)
            self._cache.pop(key, None)


# Caches to store HB and FT fonts/faces. Caches are scary, because in
# a long-running process in which many fonts come by, we don't want to
# hold/leak memory. For HB it take just 16 KB per font. For this case
# a regular dict would probably be fine. But for FT the font can take
# substantial memory (e.g. a Chinese font can take close to 1MB). This
# is why we use a temporal cache, that drops unused items after 10s.

# However, in many cases, users are probably only using a small
# subset of fonts so if there are fewer than say 5 fonts,
# we don't want to continuously evict them.
# PyGFX uses the vendored Noto fonts (up to 5 ttf files)
# And the user application may use their own 5 in a basic usecase.
# So we allow them to load approximately 20 font files without eviction


def get_hb_font(font_filename):
    ref_size = REF_GLYPH_SIZE

    blob = uharfbuzz.Blob.from_file_path(font_filename)
    face = uharfbuzz.Face(blob)
    font = uharfbuzz.Font(face)
    font.scale = ref_size, ref_size

    return blob, face, font


CACHE_HB = TemporalCache(
    lifetime=10,
    getter=get_hb_font,
    minimum_items=20,
)


def get_ft_face(font_filename):
    ref_size = REF_GLYPH_SIZE

    face = freetype.Face(font_filename)
    face.set_pixel_sizes(ref_size, ref_size)

    return face


CACHE_FT = TemporalCache(
    lifetime=10,
    getter=get_ft_face,
    minimum_items=20,
)


def shape_text_hb(text, font_filename, direction=None):
    """Shape text with Harfbuzz."""

    ref_size = REF_GLYPH_SIZE

    # Prepare buffer
    buf = uharfbuzz.Buffer()
    buf.add_str(text)
    buf.guess_segment_properties()

    is_horizontal = True
    if direction is not None:
        buf.direction = direction
        is_horizontal = direction in ("ltr", "rtl")

    # Load font, maybe from the cache
    _blob, _face, font = CACHE_HB[font_filename]

    # Shape!
    uharfbuzz.shape(font, buf)

    glyph_infos = buf.glyph_infos
    glyph_positions = buf.glyph_positions
    n_glyphs = len(glyph_infos)

    # Get glyph indices (these can be different from the text's Unicode
    # code points) and convert advances to positions.
    glyph_indices = np.empty((n_glyphs,), "u4")
    positions = np.empty((n_glyphs, 2), "f4")
    pen_x = pen_y = 0
    for i in range(n_glyphs):
        glyph_indices[i] = glyph_infos[i].codepoint
        pos = glyph_positions[i]
        positions[i] = (
            (pen_x + pos.x_offset) / ref_size,
            (pen_y + pos.y_offset) / ref_size,
        )
        pen_x += pos.x_advance
        pen_y += pos.y_advance

    # Get font extents
    font_ext = font.get_font_extents(buf.direction)

    meta = {
        "extent": abs(pen_x if is_horizontal else pen_y) / ref_size,
        "ascender": font_ext.ascender / ref_size,
        "descender": font_ext.descender / ref_size,
        "direction": buf.direction,
        "script": buf.script,
    }

    return glyph_indices, positions, meta


def shape_text_ft(text, font_filename, direction=None):
    """Shape text with FreeType.

    This function is tested but not actually used. It is provided for
    completeness, as a possible fallback. FreeType supports basic
    shaping with kerning but not much more than that (e.g. no glyph
    replacements).
    """

    # assert direction is None  # just ignore the given direction ...

    ref_size = REF_GLYPH_SIZE

    # Load font face
    face = CACHE_FT[font_filename]

    # With Freetype we simply replace each char for a glyph.
    n_glyphs = len(text)
    glyph_indices = np.array([face.get_char_index(c) for c in text], np.uint32)

    # We get the advance of each glyph (can be in font units or 16.16 format)
    advances = [
        face.get_advance(int(i), freetype.FT_LOAD_DEFAULT) for i in glyph_indices
    ]
    advances = [(x / 65536 if x > 65536 * 10 else x) for x in advances]

    # Convert advances to positions
    positions = np.empty((n_glyphs, 2), "f4")
    pen_x = 0
    prev = " "
    for i in range(n_glyphs):
        c = text[i]
        kerning = face.get_kerning(prev, c, freetype.FT_KERNING_UNSCALED)
        pen_x += kerning.x / 64
        positions[i] = pen_x / ref_size, 0
        pen_x += advances[i]
        prev = c

    meta = {
        "extent": pen_x / ref_size,
        "ascender": face.ascender / face.units_per_EM,
        "descender": face.descender / face.units_per_EM,
        "direction": "ltr",
        "script": "",
    }

    return glyph_indices, positions, meta



import wgpu  # only for flags/enums

from ....objects import Grid
from ....materials import GridMaterial

from .. import (
    register_wgpu_render_function,
    BaseShader,
    Binding,
    load_wgsl,
)


@register_wgpu_render_function(Grid, GridMaterial)
class GridShader(BaseShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        material = wobject.material

        self["thickness_space"] = material.thickness_space
        self["inf_grid"] = material.infinite
        self["draw_axis"] = material._gfx_draw_axis
        self["draw_major"] = material._gfx_draw_major
        self["draw_minor"] = material._gfx_draw_minor

    def get_bindings(self, wobject, shared):
        material = wobject.material

        bindings = {}
        bindings[0] = Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer)
        bindings[1] = Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer)
        bindings[2] = Binding("u_material", "buffer/uniform", material.uniform_buffer)
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        nverts = 30 if self["inf_grid"] else 6
        return {
            "indices": (nverts, 1),
        }

    def get_code(self):
        return load_wgsl("grid.wgsl")



"""
Animation module.

.. currentmodule:: pygfx.animation

This module contains classes for working with animations.

.. autosummary::
    :toctree: animation/
    :template: ../_templates/custom_layout.rst

    LinearInterpolant
    StepInterpolant
    QuaternionLinearInterpolant
    CubicSplineInterpolant
    KeyframeTrack
    AnimationClip

"""

# flake8: noqa

from .interpolant import (
    Interpolant,
    LinearInterpolant,
    StepInterpolant,
    QuaternionLinearInterpolant,
    CubicSplineInterpolant,
)
from .keyframe_track import KeyframeTrack
from .animation_clip import AnimationClip
from .animation_action import AnimationAction
from .animation_mixer import AnimationMixer
from .clock import Clock
